from abc import ABC, abstractmethod

from numpy.typing import NDArray, DTypeLike
from numpy import float32

from ... import Geometry


class Projector(ABC):
    """Projector for transforms of tensor fields from three-dimensional
    space to projection space.
    """

    def __init__(self,
                 geometry: Geometry):
        self._geometry = geometry
        self._geometry_hash = hash(geometry)

    @abstractmethod
    def forward(self,
                field: NDArray,
                index: NDArray[int]) -> NDArray:
        pass

    @abstractmethod
    def adjoint(self,
                projection: NDArray,
                index: NDArray[int]) -> NDArray:
        pass

    def _check_indices_kind_is_integer(self, indices: NDArray):
        if indices.dtype.kind != 'i':
            raise TypeError('Elements of indices must be of integer kind,'
                            f' but the provided indices have dtype.kind {indices.dtype.kind}!')

    @property
    def dtype(self) -> DTypeLike:
        """ The dtype input fields and projections require."""
        return float32

    @property
    def is_dirty(self) -> bool:
        """ Returns ``True`` if the system geometry has changed without
        the projection geometry having been updated. """
        return self._geometry_hash != hash(self._geometry)

    def _update(self, force_update: bool = False) -> None:
        if not (self.is_dirty or force_update):  # If not dirty, do nothing
            return
        else:
            pass  # Else, do an update as detemined by the child class definition



from typing import Any

from math import floor
from numba import config, cuda, int32, float32, float64
from numpy.typing import NDArray
import numpy as np


def _is_cuda_array(item: Any):
    """ Internal method which assists in debugging. """
    if config.ENABLE_CUDASIM:
        return False
    else:
        return cuda.is_cuda_array(item)


def john_transform_cuda(field: NDArray, projections: NDArray, unit_vector_p: NDArray,
                        unit_vector_j: NDArray, unit_vector_k: NDArray, offsets_j: NDArray,
                        offsets_k: NDArray):
    """ Frontend for performing the John transform with parallel
    GPU computing.

    Parameters
    ----------
    field
        The field to be projected, with 4 dimensions. The last index should
        have the same size as the last index of ``projections``. Can be either
        a `numpy.ndarray`, or a device array that implements the CUDA array
        interface. If a device array is given, no copying to device is needed.
    projections
        A 4-dimensional numpy array where the projections are stored.
        The first index runs over the different projection directions. Can be either
        a `numpy.ndarray`, or a device array that implements the CUDA array
        interface. If a device array is given, no copying or synchronization is needed.
    unit_vector_p
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    offsets_j
        Offsets which align projections in the direction of `j`
    offsets_k
        Offsets which align projections in the direction of `k`.

    Notes
    -----
    For mathematical details, see the :meth:<`~.john_transform.john_transform`.
    """
    # Always false for now as we are sticking to ``float32``.
    if False:
        cuda_float = float64
        numpy_float = np.float64
        cuda_x4_type = cuda.float64x4
    else:
        cuda_float = float32
        numpy_float = np.float32
        cuda_x4_type = cuda.float32x4

    # Start by transferring field/projections to GPU. (TODO: Handle better)
    if _is_cuda_array(field):
        assert field.dtype == 'float32'
        device_field = cuda.as_cuda_array(field)
    else:
        assert field.dtype == np.float32
        device_field = cuda.to_device(field)

    if _is_cuda_array(projections):
        assert projections.dtype == 'float32'
        device_projections = cuda.as_cuda_array(projections)
        transfer_projections = False
    else:
        assert projections.dtype == np.float32
        device_projections = cuda.to_device(projections)
        transfer_projections = True

    # Find zeroth, first and second directions for indexing. Zeroth is the maximal projection direction.

    # (x, y, z), (y, x, z), (z, x, y)
    direction_0_index = np.argmax(abs(unit_vector_p), axis=1).reshape(-1, 1).astype(np.int32)
    direction_1_index = 1 * (direction_0_index == 0).astype(np.int32).reshape(-1, 1).astype(np.int32)
    direction_2_index = (2 - (direction_0_index == 2)).astype(np.int32).reshape(-1, 1).astype(np.int32)

    # Step size for direction 1 and 2.
    step_sizes_1 = (np.take_along_axis(unit_vector_p, direction_1_index, 1) /
                    np.take_along_axis(unit_vector_p, direction_0_index, 1)).astype(numpy_float).ravel()
    step_sizes_2 = (np.take_along_axis(unit_vector_p, direction_2_index, 1) /
                    np.take_along_axis(unit_vector_p, direction_0_index, 1)).astype(numpy_float).ravel()

    # Shape in each of the three directions.
    dimensions_0 = np.array(field.shape, dtype=numpy_float)[direction_0_index.ravel()]
    dimensions_1 = np.array(field.shape, dtype=numpy_float)[direction_1_index.ravel()]
    dimensions_2 = np.array(field.shape, dtype=numpy_float)[direction_2_index.ravel()]

    # Correction factor for length of line when taking a one-slice step.
    distance_multipliers = np.sqrt(1.0 + step_sizes_1 ** 2 + step_sizes_2 ** 2).astype(numpy_float)

    max_index = projections.shape[0]
    max_j = projections.shape[1]
    max_k = projections.shape[2]

    # CUDA chunking and memory size constants.
    channels = int(field.shape[-1])
    blocks_j = int(32)
    blocks_k = int(32)
    blocks_angle = int(8)

    # Indices to navigate each projection. s is the surface positioning.
    k_vectors = unit_vector_k.astype(numpy_float)
    j_vectors = unit_vector_j.astype(numpy_float)
    s_vectors = (unit_vector_k * (-0.5 * max_k + offsets_k.reshape(-1, 1)) +
                 unit_vector_j * (-0.5 * max_j + offsets_j.reshape(-1, 1))).astype(numpy_float)
    s_vectors += k_vectors * 0.5 + j_vectors * 0.5

    # (0, 1, 2) if x main, (1, 0, 2) if y main, (2, 0, 1) if z main.
    direction_indices = np.stack((direction_0_index.ravel(),
                                  direction_1_index.ravel(),
                                  direction_2_index.ravel()), axis=1)

    # Bilinear interpolation over each slice.

    @cuda.jit(device=True)
    def bilinear_interpolation(field: NDArray,
                               direction_0,
                               r0: float,
                               r1: float,
                               r2: float,
                               dimensions,
                               accumulator: NDArray):
        """ Kernel for bilinear interpolation. Replaces texture interpolation."""
        if not (0 <= r1 < dimensions[1] and 0 <= r2 < dimensions[2]):
            return
        # At edges, use nearest-neighbor interpolation.
        if (1 <= int32(r1) + 1 < dimensions[1]):
            r1_weight = 1
        else:
            r1_weight = 0

        if (1 <= int32(r2) + 1 < dimensions[2]):
            r2_weight = 1
        else:
            r2_weight = 0

        if -1 < r1 < 0:
            r1_edge_weight = 1
        else:
            r1_edge_weight = 0

        if -1 < r2 < 0:
            r2_edge_weight = 1
        else:
            r2_edge_weight = 0

        # Branch should be abstracted away by compiler, but could be done with pointer arithmetic.
        if (direction_0 == 0):
            y_weight = (r1 - floor(r1)) * r1_weight
            z_weight = (r2 - floor(r2)) * r2_weight
            x = int32(r0)
            y = int32(r1) + r1_edge_weight
            z = int32(r2) + r2_edge_weight
            t = cuda_x4_type((1 - z_weight) * (1 - y_weight),
                             z_weight * y_weight,
                             (1 - z_weight) * y_weight,
                             (1 - y_weight) * z_weight)
            for i in range(accumulator.size):
                accumulator[i] += field[x, y, z][i] * t.x
                accumulator[i] += field[x, y + r1_weight, z][i] * t.z
                accumulator[i] += field[x, y, z + r2_weight][i] * t.w
                accumulator[i] += field[x, y + r1_weight, z + r2_weight][i] * t.y

        elif (direction_0 == 1):
            x_weight = (r1 - floor(r1)) * r1_weight
            z_weight = (r2 - floor(r2)) * r2_weight
            x = int32(r1) + r1_edge_weight
            y = int32(r0)
            z = int32(r2) + r2_edge_weight
            t = cuda_x4_type((1 - x_weight) * (1 - z_weight),
                             x_weight * z_weight,
                             x_weight * (1 - z_weight),
                             (1 - x_weight) * z_weight)
            for i in range(accumulator.size):
                accumulator[i] += field[x, y, z][i] * t.x
                accumulator[i] += field[x + r1_weight, y, z][i] * t.z
                accumulator[i] += field[x, y, z + r2_weight][i] * t.w
                accumulator[i] += field[x + r1_weight, y, z + r2_weight][i] * t.y

        elif (direction_0 == 2):
            x_weight = (r1 - floor(r1)) * r1_weight
            y_weight = (r2 - floor(r2)) * r2_weight
            x = int32(r1) + r1_edge_weight
            y = int32(r2) + r2_edge_weight
            z = int32(r0)
            t = cuda_x4_type((1 - x_weight) * (1 - y_weight),
                             x_weight * y_weight,
                             y_weight * (1 - x_weight),
                             (1 - y_weight) * x_weight)
            for i in range(accumulator.size):
                accumulator[i] += field[x, y, z][i] * t.x
                accumulator[i] += field[x, y + r2_weight, z][i] * t.z
                accumulator[i] += field[x + r1_weight, y, z][i] * t.w
                accumulator[i] += field[x + r1_weight, y + r2_weight, z][i] * t.y

    @cuda.jit(cache=True)
    def john_transform_inner(field: NDArray, projection: NDArray):
        """ Performs the John transform of a field. Relies on a large number
        of pre-defined constants outside the kernel body. """
        index = cuda.blockIdx.y * blocks_angle + cuda.threadIdx.y
        j = cuda.threadIdx.x + blocks_j * (
                cuda.blockIdx.x % (
                    (max_j + blocks_j - 1) // blocks_j))

        if (j >= max_j) or (index >= max_index):
            return

        start_k = blocks_k * (cuda.blockIdx.x // (
            (max_j + blocks_j - 1) // blocks_j))
        end_k = start_k + blocks_k
        if end_k > max_k:
            end_k = max_k

        if start_k >= end_k:
            return

        # Define compile-time constants.
        step_size_1 = step_sizes_1[index]
        step_size_2 = step_sizes_2[index]
        k_vectors_c = k_vectors[index]
        j_vectors_c = j_vectors[index]
        s_vectors_c = s_vectors[index]
        dimensions_0_c = dimensions_0[index]
        dimensions_1_c = dimensions_1[index]
        dimensions_2_c = dimensions_2[index]
        dimensions = cuda.local.array(3, int32)
        dimensions[0] = dimensions_0_c
        dimensions[1] = dimensions_1_c
        dimensions[2] = dimensions_2_c
        direction_indices_c = direction_indices[index]
        distance_multiplier = distance_multipliers[index]

        # Could be chunked for very asymmetric samples.
        start_slice = 0
        end_slice = start_slice + dimensions[0]
        accumulator = cuda.local.array(channels, cuda_float)

        for k in range(start_k, end_k):
            for i in range(channels):
                accumulator[i] = 0.

            fj = cuda_float(j)
            fk = cuda_float(k)

            # Initial coordinates of frame.
            start_position_0 = (s_vectors_c[direction_indices_c[0]] +
                                fj * j_vectors_c[direction_indices_c[0]] +
                                fk * k_vectors_c[direction_indices_c[0]])
            start_position_1 = (s_vectors_c[direction_indices_c[1]] +
                                fj * j_vectors_c[direction_indices_c[1]] +
                                fk * k_vectors_c[direction_indices_c[1]])
            start_position_2 = (s_vectors_c[direction_indices_c[2]] +
                                fj * j_vectors_c[direction_indices_c[2]] +
                                fk * k_vectors_c[direction_indices_c[2]])

            # Centering w.r.t volume.
            centering_step_1 = start_position_1 - step_size_1 * start_position_0
            centering_step_2 = start_position_2 - step_size_2 * start_position_0

            position_0 = cuda_float(start_slice)
            position_1 = step_size_1 * (cuda_float(start_slice) - 0.5 * dimensions[0] + 0.5) + \
                centering_step_1 + 0.5 * dimensions[1]
            position_2 = step_size_2 * (cuda_float(start_slice) - 0.5 * dimensions[0] + 0.5) + \
                centering_step_2 + 0.5 * dimensions[2]

            for i in range(start_slice, end_slice):
                bilinear_interpolation(field, direction_indices_c[0],
                                       position_0, position_1, position_2, dimensions, accumulator)
                position_0 += 1.0
                position_1 += step_size_1
                position_2 += step_size_2

            for i in range(channels):
                projection[index, j, k, i] += accumulator[i] * distance_multiplier

    # Launching of kernel.
    bpg = (((max_j + blocks_j - 1) // blocks_j) *
           ((max_k + blocks_k - 1) // blocks_k),
           ((projections.shape[0] + blocks_angle - 1) // blocks_angle))
    tpb = (blocks_j, blocks_angle)
    john_transform_inner[bpg, tpb](device_field,
                                   device_projections)
    if transfer_projections:
        device_projections.copy_to_host(projections)


def john_transform_adjoint_cuda(field: NDArray, projections: NDArray, unit_vector_p: NDArray,
                                unit_vector_j: NDArray, unit_vector_k: NDArray, offsets_j: NDArray,
                                offsets_k: NDArray):
    """ Frontend for performing the adjoint of the John transform with parallel
    GPU computing.

    Parameters
    ----------
    field
        The field into which the adjoint is projected, with 4 dimensions. The last index should
        have the same size as the last index of ``projections``. Can be either
        a `numpy.ndarray`, or a device array that implements the CUDA array
        interface. If a device array is given, no copying to device is needed.
    projections
        The projections from which the adjoint is calculated.
        The first index runs over the different projection directions. Can be either
        a `numpy.ndarray`, or a device array that implements the CUDA array
        interface. If a device array is given, no copying or synchronization is needed.
    unit_vector_p
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    offsets_j
        Offsets which align projections in the direction of `j`
    offsets_k
        Offsets which align projections in the direction of `k`.

    Notes
    -----
    For mathematical details, see the :meth:<`~.john_transform.john_transform_adjoint`.
    """
    # Always false as we are sticking to float32
    if False:
        cuda_float = float64
        numpy_float = np.float64
        cuda_x4_type = cuda.float64x4
    else:
        cuda_float = float32
        numpy_float = np.float32
        cuda_x4_type = cuda.float32x4

    if _is_cuda_array(field):
        assert field.dtype == 'float32'
        device_field = cuda.as_cuda_array(field)
        transfer_field = False
    else:
        assert field.dtype == np.float32
        device_field = cuda.to_device(field)
        transfer_field = True

    if _is_cuda_array(projections):
        assert projections.dtype == 'float32'
        device_projections = cuda.as_cuda_array(projections)
    else:
        assert projections.dtype == np.float32
        device_projections = cuda.to_device(projections)

    max_j = projections.shape[1]
    max_k = projections.shape[2]
    # CUDA chunking and memory size constants.
    blocks_z = int(1)
    # blocks_angle = int(512)  # Not used for now.
    blocks_x = int(16)
    blocks_y = int(32)

    # Projection and frame vectors. s for positioning the frame.
    p_vectors = unit_vector_p.astype(numpy_float)
    k_vectors = unit_vector_k.astype(numpy_float)
    j_vectors = unit_vector_j.astype(numpy_float)
    s_vectors = (unit_vector_k * (-0.5 * max_k + offsets_k.reshape(-1, 1)) +
                 unit_vector_j * (-0.5 * max_j + offsets_j.reshape(-1, 1))).astype(numpy_float)

    # Translate volume steps to normalized frame steps. Can add support for non-square voxels.
    vector_norm = np.einsum('...i, ...i', p_vectors, np.cross(j_vectors, k_vectors))
    norm_j = -np.cross(p_vectors, k_vectors) / vector_norm[..., None]
    norm_k = np.cross(p_vectors, j_vectors) / vector_norm[..., None]
    norm_offset_j = -np.einsum('...i, ...i', p_vectors, np.cross(s_vectors, k_vectors)) / vector_norm
    norm_offset_k = np.einsum('...i, ...i', p_vectors, np.cross(s_vectors, j_vectors)) / vector_norm

    channels = field.shape[-1]

    @cuda.jit(device=True)
    def bilinear_interpolation_projection(projection: NDArray,
                                          r0: float,
                                          r1: float,
                                          r2: float,
                                          accumulator: NDArray):
        if not (-1 <= r1 < max_j and -1 <= r2 < max_k):
            return
        # At edges, use nearest-neighbor interpolation.
        if (0 <= int32(r1) + 1 < max_j):
            r1_weight = 1
        else:
            r1_weight = 0

        if (0 <= int32(r2) + 1 < max_k):
            r2_weight = 1
        else:
            r2_weight = 0

        if -1 < r1 < 0:
            r1_edge_weight = 1
        else:
            r1_edge_weight = 0

        if -1 < r2 < 0:
            r2_edge_weight = 1
        else:
            r2_edge_weight = 0

        y_weight = (r1 - floor(r1)) * r1_weight
        z_weight = (r2 - floor(r2)) * r2_weight
        x = int32(r0)
        y = int32(r1) + r1_edge_weight
        z = int32(r2) + r2_edge_weight
        t = cuda_x4_type((1 - z_weight) * (1 - y_weight),
                         z_weight * y_weight,
                         (1 - z_weight) * y_weight,
                         (1 - y_weight) * z_weight)
        for i in range(channels):
            accumulator[i] += projection[x, y, z][i] * t.x
            accumulator[i] += projection[x, y + r1_weight, z][i] * t.z
            accumulator[i] += projection[x, y, z + r2_weight][i] * t.w
            accumulator[i] += projection[x, y + r1_weight, z + r2_weight][i] * t.y

    @cuda.jit(cache=True)
    def john_transform_adjoint_inner(field: NDArray, projection: NDArray, start_index: int, stop_index: int):
        """ Performs the John transform of a field. Relies on a large number
        of pre-defined constants outside the kernel body. """
        # Indexing of volume coordinates.
        x = cuda.threadIdx.x + blocks_x * (cuda.blockIdx.x % (
                       (field.shape[0] + blocks_x - 1) // blocks_x))
        y = cuda.threadIdx.y + blocks_y * (cuda.blockIdx.x // (
                       (field.shape[0] + blocks_x - 1) // blocks_x))
        if (x >= field.shape[0]) or (y >= field.shape[1]):
            return

        # Stride in z.
        z = cuda.blockIdx.y * blocks_z
        if z >= field.shape[2]:
            return

        # Center of voxel and coordinate system.
        fx = x - 0.5 * field.shape[0]
        fy = y - 0.5 * field.shape[1]
        fz = z - 0.5 * field.shape[2]

        z_acc = cuda.local.array(channels, cuda_float)
        for j in range(channels):
            z_acc[j] = 0.0

        # Compile time constants
        norm_j_c = norm_j[start_index:stop_index]
        norm_k_c = norm_k[start_index:stop_index]
        norm_offset_j_c = norm_offset_j[start_index:stop_index]
        norm_offset_k_c = norm_offset_k[start_index:stop_index]

        for a in range(start_index, stop_index):
            # Center with respect to projection.
            fj = norm_offset_j_c[a] + fx * norm_j_c[a][0] + fy * norm_j_c[a][1] + fz * norm_j_c[a][2]
            fk = norm_offset_k_c[a] + fx * norm_k_c[a][0] + fy * norm_k_c[a][1] + fz * norm_k_c[a][2]
            bilinear_interpolation_projection(projection,
                                              a, fj, fk, z_acc)
        for i in range(channels):
            field[x, y, z, i] += z_acc[i]

    bpg = (((field.shape[0] + blocks_x - 1) // blocks_x) *
           ((field.shape[1] + blocks_y - 1) // blocks_y),
           ((field.shape[2] + blocks_z - 1) // blocks_z))
    tpb = (blocks_x, blocks_y)
    # Last two arguments hardcoded for now, until I work out how to do chunking efficiently.
    john_transform_adjoint_inner[bpg, tpb](device_field,
                                           device_projections,
                                           0, projections.shape[0])

    if transfer_field:
        device_field.copy_to_host(field)



# -*- coding: utf-8 -*-
"""
Main module of the mumott package.
"""

import logging
from .core.numba_setup import numba_setup
from .data_handling.geometry import Geometry
from .core.probed_coordinates import ProbedCoordinates
from .data_handling.data_container import DataContainer

__project__ = 'mumott'
__description__ = 'A module for analyzing tensor tomography experiments via Python'
__copyright__ = '2023'
__license__ = 'Mozilla Public License 2.0 (MPL 2.0)'
__version__ = '0.4'
__maintainer__ = 'The mumott developers team'
__status__ = 'Beta'
__url__ = 'https://mumott.org/'

__all__ = [
    'Geometry',
    'ProbedCoordinates',
    'DataContainer'
]

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
numba_setup()



# -*- coding: utf-8 -*-


from .saxs_projector_astra import SAXSProjectorAstra
from .saxs_projector_cuda import SAXSProjectorCUDA
from .saxs_projector_numba import SAXSProjectorNumba

__all__ = [
    'SAXSProjectorAstra',
    'SAXSProjectorCUDA',
    'SAXSProjectorNumba',
]



import pytest # noqa
from io import StringIO
import logging
import numba
import numpy as np
from mumott.data_handling import DataContainer
from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjectorNumba
from mumott.methods.functionals import SIGTT
from mumott.optimization.loss_functions import SquaredLoss
from mumott.optimization.optimizers import LBFGS


def test_minimal_pipeline():
    numba.config.NUMBA_NUM_THREADS = 1
    numba.set_num_threads(1)

    # check that the functionality for deleting frames works
    data_container = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    stack = data_container.stack
    assert len(stack) == 1
    del stack[0]
    assert len(stack) == 0

    # minimal check for append functionality
    dc1 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    stack1 = dc1.stack
    dc2 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    stack2 = dc2.stack
    assert len(stack1) == 1
    frame = stack2[0]
    del stack2[0]
    stack1.append(frame)
    assert len(stack1) == 2

    # minimal check for setitem functionality
    dc1 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    dc2 = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    frame = dc2.stack[0]
    frame.j_offset = 1.243

    assert dc2.geometry.j_offsets[0] == 1.243

    dc2.geometry.k_offsets[0] = 4.214

    assert frame.k_offset == 4.214

    assert dc2.stack.geometry[0] == frame.geometry

    del dc2.stack[0]

    assert frame.j_offset == 1.243
    assert frame.k_offset == 4.214
    frame.data[0][0] = [1, 2, 3]
    dc1.stack[0] = frame

    assert np.all(dc1.stack[0].data[0][0] == [1, 2, 3])
    assert dc1.stack[0].j_offset == 1.243
    assert dc1.stack[0].k_offset == 4.214
    assert dc1.stack[0].geometry == dc1.stack.geometry[0]

    # load data container and check its basic output
    data_container = DataContainer(data_path='tests/test_half_circle.h5', data_type='h5')
    s = str(data_container)
    assert 'DataContainer' in s
    assert 'Corrected for transmission' in s
    assert 'Corrected for transmission' in s
    s = data_container._repr_html_()
    assert 'DataContainer' in s
    assert 'Corrected for transmission' in s
    assert 'Corrected for transmission' in s

    s = str(data_container.geometry)
    assert 'Geometry' in s
    s = data_container.geometry._repr_html_()
    assert 'Geometry' in s
    s = str(data_container.stack)
    assert 'Stack' in s
    s = data_container.stack._repr_html_()
    assert 'Stack' in s

    s = str(data_container.stack[0])
    assert 'Frame' in s
    assert 'diode' in s
    assert 'dcf160' in s
    s = data_container.stack[0]._repr_html_()
    assert 'Frame' in s
    assert 'diode' in s
    assert 'dcf160' in s

    bs = SphericalHarmonics(ell_max=6)
    pr = SAXSProjectorNumba(data_container.geometry)
    meth = SIGTT(data_container, bs, pr)
    lf = SquaredLoss(meth)
    optimizer = LBFGS(lf, maxiter=5)
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    # - send logging to stream
    f = StringIO()
    logging.basicConfig(stream=f, level=logging.INFO)
    # run optimization
    optimizer.optimize()

    # check that coefficients are correct
    coeffs = meth.coefficients
    assert coeffs.size == 1792
    reference_coeffs = np.array(
        [0.87601399, 0.08113027, -0.00716787,  0.48052269, -0.15047219, 0.84963459,
         0.13741241, -0.01485878, 0.05635246, -0.00441374,  0.31243902, -0.09265587,
         0.5901497, -0.10334509, 0.71296326, 0.18458887, -0.01809086, 0.10050705,
         -0.01510339, 0.04192086, -0.00137633,  0.26808559, -0.02889264, 0.43901513,
         -0.10504635, 0.52148008, -0.07457155,  0.62865195, 0.88099492, 0.07854443,
         -0.01004529, 0.53877781, -0.21087649,  0.82255451, 0.1297739, -0.02188681,
         0.0632176, -0.00524503, 0.29931401, -0.11010673])

    print(coeffs.ravel()[:40])
    assert np.allclose(coeffs.ravel()[:40], reference_coeffs)



import pytest
import logging
import numpy as np
from numpy import array

from mumott.methods.basis_sets import SphericalHarmonics
from mumott import ProbedCoordinates

LOGGER = logging.getLogger(__name__)

output_dictionary = {'basis_set': {
    'name': 'SphericalHarmonics',
    'coefficients': array([[10, 15,  7,  9,  3,  6]]),
    'ell_max': 2,
    'ell_indices': array([0, 2, 2, 2, 2, 2]),
    'emm_indices': array([0, -2, -1,  0,  1,  2]),
    'projection_matrix': array([[[[1., 0., -0., -1.11803399, -0., 1.93649167]]]]),
    'hash': '0x1fc6a9246a9f8dfb'},
    'spherical_functions':
    {'means': array([10]),
     'variances': array([400]),
     'eigenvectors': array([[[0.53554753,  0.64567032, -0.54433324],
                             [-0.83389412,  0.30245493, -0.46167263],
                             [0.13345204, -0.70116393, -0.70039967]]]),
     'r2_tensors': array([[3, -43,  40,  27,  11,  58]]),
     'tensor_to_matrix_indices': [[0, 5, 4], [5, 1, 3], [4, 3, 2]],
     'eigenvalues': array([[-84.56998002,  18.22384166,  66.34613836]]),
     'main_orientations': array([[[0.53554753],
                                  [-0.83389412],
                                  [0.13345204]]]),
     'main_orientation_symmetries': array([-0.43097661]),
     'normalized_standard_deviations': array([2]),
     'power_spectra': array([[100., 400.]]),
     'power_spectra_ell': array([0, 2])}}

default_coeffs = array((10, 15, 7, 9, 3, 6)).reshape(1, 6)

default_pc = ProbedCoordinates()

probed_coordinate_list = [ProbedCoordinates(vector=np.array((0., 1., 0)).reshape(1, 1, 1, 3)),
                          ProbedCoordinates(vector=np.array((0., 0.7, -0.7)).reshape(1, 1, 1, 3))]

pc_vector_list = [np.array((0., 1., 0.)).reshape(1, 1, 1, 3),
                  np.array((0., 0.7, -0.7)).reshape(1, 1, 1, 3)]

pc_hash_list = ['642022', '173557']

ell_max_list = [2, 4, 6]
expected_hashes = ['228970', '791168', '374938']


@pytest.mark.parametrize('ell_max, expected_ell, expected_emm', [(0, [0], [0]),
                                                                 (2,
                                                                  [0, 2, 2, 2, 2, 2],
                                                                  [0, -2, -1, 0, 1, 2])])
def test_ell_max(ell_max, expected_ell, expected_emm):
    s = SphericalHarmonics(default_pc, ell_max)
    assert s.ell_max == ell_max
    assert np.allclose(s.ell_indices, expected_ell)
    assert np.allclose(s.emm_indices, expected_emm)


@pytest.mark.parametrize('ell_max,expected_matrix', [(0, [[[1]]]),
                                                     (2, [[[[1., 0., -0., -1.11803399, -0., 1.93649167]]]])])
def test_projection_matrix(ell_max, expected_matrix):
    s = SphericalHarmonics(default_pc, ell_max)
    assert np.allclose(s.projection_matrix, expected_matrix)


def test_r2_ell0(caplog):
    s = SphericalHarmonics(default_pc, 0)
    caplog.set_level(logging.INFO)
    r2, eve, ev = s._get_rank2_tensor_analysis(default_coeffs[..., 0])
    assert 'ell_max < 2' in caplog.text
    assert np.allclose(ev[..., 0], -1)
    assert np.allclose(ev[..., 1], 0)
    assert np.allclose(ev[..., 2], 1)


@pytest.mark.parametrize('ell_max, coeffs, expected_dictionary', [
    (2, default_coeffs, output_dictionary)])
def test_output(ell_max, coeffs, expected_dictionary):
    s = SphericalHarmonics(default_pc, ell_max)
    d = s.get_output(coeffs)
    for key, value in d['basis_set'].items():
        if type(value) is str:
            assert expected_dictionary['basis_set'][key] == value
        else:
            assert np.allclose(expected_dictionary['basis_set'][key], value)
    for key, value in d['spherical_functions'].items():
        if type(value) is str:
            assert expected_dictionary['spherical_functions'][key] == value
        elif key == 'eigenvectors':
            assert np.allclose(abs(np.einsum('...ik, ...ik -> ...k',
                                   expected_dictionary['spherical_functions'][key],
                                   value)), 1.)
        elif key == 'main_orientations':
            assert np.allclose((abs(np.einsum('...i, ...i -> ...',
                               expected_dictionary['spherical_functions'][key].squeeze(),
                               value.squeeze()))), 1.)
        else:
            assert np.allclose(expected_dictionary['spherical_functions'][key], value)


@pytest.mark.parametrize('ell_max, coeffs, expected_result', [
    (2, default_coeffs, array([[11.55664414]]))])
def test_forward(ell_max, coeffs, expected_result):
    s = SphericalHarmonics(default_pc, ell_max)
    r = s.forward(default_coeffs, np.array((0,)))
    assert np.allclose(r, expected_result)


@pytest.mark.parametrize('ell_max, coeffs, expected_result', [
    (2, array([[11.55664414]]), array([[[11.55664414, 0., 0., -12.92072094, 0., 22.37934515]]]))])
def test_gradient(ell_max, coeffs, expected_result):
    s = SphericalHarmonics(default_pc, ell_max)
    r = s.gradient(coeffs)
    assert np.allclose(r, expected_result)
    r = s.gradient(coeffs, np.array([0]))
    assert np.allclose(r, expected_result[0])


def test_covariances():
    s = SphericalHarmonics(default_pc, 2)
    p = s.get_inner_product(default_coeffs, default_coeffs)
    assert p == 500
    p = s.get_inner_product(default_coeffs, default_coeffs, spectral_moments=[2])
    assert p == 400
    p = s.get_inner_product(default_coeffs, default_coeffs, resolve_spectrum=True)
    assert np.allclose(p, (100, 400))


@pytest.mark.parametrize('pc, pc_vec, pc_hash', [
                         t for t in zip(probed_coordinate_list,
                                        pc_vector_list,
                                        pc_hash_list)])
def test_probed_coordinates(pc, pc_vec, pc_hash):
    s = SphericalHarmonics(pc)
    assert np.allclose(s.probed_coordinates.vector, pc_vec)
    assert str(hash(s.probed_coordinates))[:6] == pc_hash
    s = SphericalHarmonics()
    s.probed_coordinates.vector = pc_vec
    assert s.is_dirty
    s._update()
    assert not s.is_dirty
    assert np.allclose(s.probed_coordinates.vector, pc.vector)
    assert str(hash(s.probed_coordinates))[:6] == pc_hash
    assert str(s._probed_coordinates_hash)[:6] == pc_hash


@pytest.mark.parametrize('ell_max, expected_hash', [t for t in zip(ell_max_list, expected_hashes)])
def test_hash(ell_max, expected_hash):
    s = SphericalHarmonics(ell_max=ell_max)
    assert str(hash(s))[:6] == expected_hash
    old_pm = s._projection_matrix.copy()
    where = abs(s._projection_matrix) > 0
    s._projection_matrix[where] += abs(s._projection_matrix[where]).min() * 1e-8
    assert np.any(old_pm != s._projection_matrix)
    assert np.allclose(old_pm, s._projection_matrix)
    assert str(hash(s))[:6] == expected_hash
    s._projection_matrix = s._projection_matrix + s._projection_matrix.min() * 1e-4
    assert str(hash(s))[:6] != expected_hash


def test_str():
    s = SphericalHarmonics(ell_max=6)
    string = str(s)
    assert '5340c5' in string


def test_html():
    s = SphericalHarmonics(ell_max=6)
    html = s._repr_html_()
    assert '5340c5' in html


def test_set_ell_max():
    s = SphericalHarmonics(ell_max=6)
    s.ell_max = 8
    assert s.ell_max == 8
    with pytest.raises(ValueError, match='even'):
        s.ell_max = -5

    with pytest.raises(ValueError, match='even'):
        s.ell_max = 3

    with pytest.raises(ValueError, match='even'):
        s.ell_max = 1.5


def test_dict():
    s = SphericalHarmonics(ell_max=6)
    d = dict(s)
    assert str(d['hash'])[:6] == '5340c5'



from __future__ import annotations
import logging
import tarfile
import tempfile
import json
import os
import codecs
import numpy as np
from typing import List, NamedTuple, Union
from numpy.typing import NDArray
from ..core.hashing import list_to_hash

logger = logging.getLogger(__name__)


class GeometryTuple(NamedTuple):
    """Tuple for passing and returning frame-wise geometry information.
    This is a helper class used by :class:`Geometry`.

    Attributes
    ----------
    rotation
        Rotation matrix.
    j_offset
        Offset to align projection in the direction j.
    k_offset
        Offset to align projection in the direction k.
    """
    rotation: NDArray[np.float64] = np.eye(3, dtype=np.float64)
    j_offset: np.float64 = np.float64(0)
    k_offset: np.float64 = np.float64(0)

    def __hash__(self) -> int:
        to_hash = [self.rotation.ravel(), self.j_offset, self.k_offset]
        return int(list_to_hash(to_hash), 16)

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += ['GeometryTuple'.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, precision=5, linewidth=60, edgeitems=1):
            ss = ', '.join([f'{r}' for r in self.rotation])
            s += ['{:18} : {}'.format('rotation', ss)]
            s += ['{:18} : {}'.format('j_offset', self.j_offset)]
            s += ['{:18} : {}'.format('k_offset', self.k_offset)]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += ['<h3>GeometryTuple</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">Rotation</td>']
            s += [f'<td>{self.rotation.shape}</td><td>{self.rotation}</td></tr>']
            s += ['<tr><td style="text-align: left;">j_offset</td>']
            s += [f'<td>{1}</td><td>{self.j_offset}</td></tr>']
            s += ['<tr><td style="text-align: left;">k_offset</td>']
            s += [f'<td>{1}</td><td>{self.k_offset}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)


class Geometry:
    """ Stores information about the system geometry.
    Instances of this class are used by :class:`DataContainer`
    and :class:`Stack <mumott.data_handling.stack.Stack>` to maintain geometry information.
    They can be stored as a file using the :meth:`write` method.
    This allows one to (re)create a :class:`Geometry` instance from an earlier
    and overwrite the geometry information read by :class:`DataContainer`.
    This is useful, for example, in the context of alignment.

    Parameters
    ----------
    filename
        Name of file from which to read geometry information.
        Defaults to ``None``, in which case the instance is created with
        default parameters.
    """
    def __init__(self, filename: str = None):
        self._rotations = []
        self._j_offsets = []
        self._k_offsets = []
        self.p_direction_0 = np.array([0, 1, 0]).astype(np.float64)
        self.j_direction_0 = np.array([1, 0, 0]).astype(np.float64)
        self.k_direction_0 = np.array([0, 0, 1]).astype(np.float64)
        self.detector_direction_origin = np.array([1, 0, 0]).astype(np.float64)
        self.detector_direction_positive_90 = np.array([0, 0, 1]).astype(np.float64)
        self.projection_shape = np.array([0, 0]).astype(np.int32)
        self.volume_shape = np.array([0, 0, 0]).astype(np.int32)
        self.detector_angles = np.array([]).astype(np.float64)
        if filename is not None:
            self.read(filename)

    def write(self, filename: str) -> None:
        """Method for writing the current state of a :class:`Geometry` instance to file.

        Parameters
        ----------
        filename
            Name of output file.
        """
        to_write = dict(_rotations=self.rotations_as_array.tolist(),
                        _j_offsets=self._j_offsets,
                        _k_offsets=self._k_offsets,
                        p_direction_0=self.p_direction_0.tolist(),
                        j_direction_0=self.j_direction_0.tolist(),
                        k_direction_0=self.k_direction_0.tolist(),
                        detector_direction_origin=self.detector_direction_origin.tolist(),
                        detector_direction_positive_90=self.detector_direction_positive_90.tolist(),
                        projection_shape=self.projection_shape.tolist(),
                        volume_shape=self.volume_shape.tolist(),
                        detector_angles=self.detector_angles.tolist(),
                        checksum=[hash(self)])
        with tarfile.open(name=filename, mode='w') as tar_file:
            for key, item in to_write.items():
                temp_file = tempfile.NamedTemporaryFile(delete=False)
                temp_file.close()
                with codecs.open(temp_file.name, 'w', encoding='utf-8') as tf:
                    json.dump(item, tf)
                    tf.flush()
                with open(temp_file.name, 'rb') as tt:
                    tar_info = tar_file.gettarinfo(arcname=key, fileobj=tt)
                    tar_file.addfile(tar_info, tt)
                os.remove(temp_file.name)

    def read(self, filename: str) -> None:
        """Method for reading the current state of a :class:`Geometry` instance from file.

        Parameters
        ----------
        filename
            Name of input file.
        """
        to_read = ['_rotations',
                   '_j_offsets',
                   '_k_offsets',
                   'p_direction_0',
                   'j_direction_0',
                   'k_direction_0',
                   'detector_direction_origin',
                   'detector_direction_positive_90',
                   'projection_shape',
                   'volume_shape',
                   'detector_angles',
                   'checksum']
        with tarfile.open(name=filename, mode='r') as tar_file:
            for key in to_read:
                temp_file = tempfile.NamedTemporaryFile(delete=False)
                temp_file.write(tar_file.extractfile(key).read())
                temp_file.close()
                with codecs.open(temp_file.name, 'r', encoding='utf-8') as file:
                    text = file.read()
                data_as_list = json.loads(text)
                if key == 'checksum':
                    checksum = data_as_list[0]
                elif key == 'rotations':
                    self._rotations = []
                    for entry in data_as_list:
                        self._rotations.append(entry)
                elif key in ('_j_offsets', '_k_offsets'):
                    setattr(self, key, data_as_list)
                else:
                    setattr(self, key, np.array(data_as_list))
        if checksum != hash(self):
            logger.warning(f'Checksum does not match! Checksum is {checksum},'
                           f' but hash(self) is {hash(self)}. This may be due to'
                           ' version differences, but please proceed with caution!')

    def append(self, value: GeometryTuple) -> None:
        """ Appends frame-wise data handed via a
        :class:`GeometryTuple <mumott.data_handling.geometry.GeometryTuple>`. """
        self._rotations.append(value.rotation)
        self._j_offsets.append(value.j_offset)
        self._k_offsets.append(value.k_offset)

    def insert(self, key: int, value: GeometryTuple) -> None:
        """ Inserts frame-wise data handed via a
        :class:`GeometryTuple <mumott.data_handling.geometry.GeometryTuple>`. """
        self._rotations.insert(key, value.rotation)
        self._j_offsets.insert(key, value.j_offset)
        self._k_offsets.insert(key, value.k_offset)

    def __setitem__(self, key: int, value: GeometryTuple) -> None:
        """ Sets frame-wise data handed via a :class:`GeometryTuple`."""
        self._rotations[key] = value.rotation
        self._j_offsets[key] = value.j_offset
        self._k_offsets[key] = value.k_offset

    def __getitem__(self, key: int) -> GeometryTuple:
        """ Returns frame-wise data handed via a :class:`GeometryTuple`."""
        return GeometryTuple(rotation=self._rotations[key],
                             j_offset=self._j_offsets[key],
                             k_offset=self._k_offsets[key])

    def __delitem__(self, key: int) -> None:
        del self._rotations[key]
        del self._j_offsets[key]
        del self._k_offsets[key]

    @property
    def rotations_as_array(self) -> NDArray:
        """ Rotation matrices corresponding to each frame of data as an array."""
        if len(self) == 0:
            return np.array([])
        return np.stack([r for r in self._rotations], axis=0)

    @property
    def rotations(self) -> List[NDArray[np.float64]]:
        """ Rotation matrices corresponding to each frame of data."""
        return self._rotations

    @rotations.setter
    def rotations(self, value: Union[list, NDArray]) -> None:
        self._rotations = [r for r in value]

    @property
    def j_offsets(self) -> List[np.float64]:
        """Offsets to align projection in the direction j as an array."""
        return self._j_offsets

    @property
    def j_offsets_as_array(self) -> NDArray:
        """Offsets to align projection in the direction j as an array."""
        if len(self._j_offsets) == 0:
            return np.array([])
        return np.stack([j for j in self._j_offsets], axis=0)

    @j_offsets.setter
    def j_offsets(self, value: Union[List[np.float64], NDArray]) -> None:
        self._j_offsets = [j for j in value]

    @property
    def k_offsets(self) -> List[np.float64]:
        """Offsets to align projection in the direction k."""
        return self._k_offsets

    @property
    def k_offsets_as_array(self) -> NDArray:
        """Offsets to align projection in the direction k as an array."""
        if len(self._k_offsets) == 0:
            return np.array([])
        return np.stack([k for k in self._k_offsets], axis=0)

    @property
    def hash_rotations(self) -> str:
        """ A sha1 hash of :attr:`rotations_as_array`. """
        return list_to_hash([self.rotations_as_array])

    @property
    def hash_j_offsets(self) -> str:
        """ A sha1 hash of :attr:`j_offsets_as_array`. """
        return list_to_hash([self.j_offsets_as_array])

    @property
    def hash_k_offsets(self) -> str:
        """ A sha1 hash of :attr:`k_offsets_as_array`. """
        return list_to_hash([self.k_offsets_as_array])

    @k_offsets.setter
    def k_offsets(self, value: Union[List[np.float64], NDArray]) -> None:
        self._k_offsets = [k for k in value]

    def __hash__(self) -> int:
        to_hash = [self.rotations_as_array,
                   self.j_offsets_as_array,
                   self.k_offsets_as_array,
                   self.p_direction_0, self.j_direction_0, self.k_direction_0,
                   self.detector_direction_origin, self.detector_direction_positive_90,
                   self.projection_shape, self.volume_shape, self.detector_angles]
        return int(list_to_hash(to_hash), 16)

    def __len__(self) -> int:
        return len(self._rotations)

    def _get_str_representation(self, max_lines: int = 25) -> str:
        """ Retrieves a string representation of the object with specified
        maximum number of lines.

        Parameters
        ----------
        max_lines
            The maximum number of lines to return.
        """
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += ['Geometry'.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=3, edgeitems=1, precision=3, linewidth=60):
            s += ['{:18} : {}'.format('hash_rotations',
                  self.hash_rotations[:6])]
            s += ['{:18} : {}'.format('hash_j_offsets',
                  self.hash_j_offsets[:6])]
            s += ['{:18} : {}'.format('hash_k_offsets',
                  self.hash_k_offsets[:6])]
            s += ['{:18} : {}'.format('p_direction_0', self.p_direction_0)]
            s += ['{:18} : {}'.format('j_direction_0', self.j_direction_0)]
            s += ['{:18} : {}'.format('k_direction_0', self.k_direction_0)]
            s += ['{:18} : {}'.format('detector_direction_origin', self.detector_direction_origin)]
            s += ['{:18} : {}'.format('detector_direction_positive_90', self.detector_direction_positive_90)]
            s += ['{:18} : {}'.format('projection_shape', self.projection_shape)]
            s += ['{:18} : {}'.format('volume_shape', self.volume_shape)]
            s += ['{:18} : {}'.format('detector_angles', self.detector_angles)]
        s += ['-' * wdt]
        truncated_s = []
        leave_loop = False
        while not leave_loop:
            line = s.pop(0).split('\n')
            for split_line in line:
                if split_line != '':
                    truncated_s += [split_line]
                if len(truncated_s) > max_lines - 2:
                    if split_line != '...':
                        truncated_s += ['...']
                    if split_line != ('=' * wdt):
                        truncated_s += ['=' * wdt]
                    leave_loop = True
                    break
            if len(s) == 0:
                leave_loop = True
        return '\n'.join(truncated_s)

    def __str__(self) -> str:
        return self._get_str_representation()

    def _get_html_representation(self, max_lines: int = 25) -> str:
        """ Retrieves an html representation of the object with specified
        maximum number of lines.

        Parameters
        ----------
        max_lines
            The maximum number of lines to return.
        """
        s = []
        s += ['<h3>Geometry</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=3, edgeitems=1, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">rotations</td>']
            s += [f'<td>{len(self.rotations)}</td>'
                  f'<td>{self.hash_rotations[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">j_offsets</td>']
            s += [f'<td>{len(self.j_offsets)}</td>'
                  f'<td>{self.hash_j_offsets[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">k_offsets</td>']
            s += [f'<td>{len(self.k_offsets)}</td>'
                  f'<td>{self.hash_k_offsets[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">p_direction_0</td>']
            s += [f'<td>{len(self.p_direction_0)}</td><td>{self.p_direction_0}</td></tr>']
            s += ['<tr><td style="text-align: left;">j_direction_0</td>']
            s += [f'<td>{len(self.j_direction_0)}</td><td>{self.j_direction_0}</td></tr>']
            s += ['<tr><td style="text-align: left;">k_direction_0</td>']
            s += [f'<td>{len(self.k_direction_0)}</td><td>{self.k_direction_0}</td></tr>']
            s += ['<tr><td style="text-align: left;">detector_direction_origin</td>']
            s += [f'<td>{len(self.detector_direction_origin)}</td>'
                  f'<td>{self.detector_direction_origin}</td></tr>']
            s += ['<tr><td style="text-align: left;">detector_direction_positive_90</td>']
            s += [f'<td>{len(self.detector_direction_positive_90)}</td>'
                  f'<td>{self.detector_direction_positive_90}</td></tr>']
            s += ['<tr><td style="text-align: left;">projection_shape</td>']
            s += [f'<td>{len(self.projection_shape)}</td><td>{self.projection_shape}</td></tr>']
            s += ['<tr><td style="text-align: left;">volume_shape</td>']
            s += [f'<td>{len(self.volume_shape)}</td><td>{self.volume_shape}</td></tr>']
            s += ['<tr><td style="text-align: left;">detector_angles</td>']
            s += [f'<td>{len(self.detector_angles)}</td><td>{self.detector_angles}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        truncated_s = []
        line_count = 0
        leave_loop = False
        while not leave_loop:
            line = s.pop(0).split('\n')
            for split_line in line:
                truncated_s += [split_line]
                if '</tr>' in split_line:
                    line_count += 1
                    # Catch if last line had ellipses
                    last_tr = split_line
                if line_count > max_lines - 1:
                    if last_tr != '<tr><td style="text-align: left;">...</td></tr>':
                        truncated_s += ['<tr><td style="text-align: left;">...</td></tr>']
                    truncated_s += ['</tbody>']
                    truncated_s += ['</table>']
                    leave_loop = True
                    break
            if len(s) == 0:
                leave_loop = True
        return '\n'.join(truncated_s)

    def _repr_html_(self) -> str:
        return self._get_html_representation()



# -*- coding: utf-8 -*-

from .orientation_image_mapper import OrientationImageMapper
from .projection_viewer import ProjectionViewer
from .saving import dict_to_h5

__all__ = [
    'OrientationImageMapper',
    'ProjectionViewer',
    'dict_to_h5'
]



try:
    import astra
except ImportError:
    raise ImportError('ASTRA is not installed. SAXSProjectorAstra is not available.')

import logging
import numpy as np
from numpy.typing import NDArray
from ... import Geometry
from ...core.hashing import list_to_hash
from .base_projector import Projector


logger = logging.getLogger(__name__)


class SAXSProjectorAstra(Projector):
    """
    Projector for transforms of tensor fields from three-dimensional space
    to projection space.

    This version utilizes the high-performance toolbox ASTRA (https://www.astra-toolbox.com/)
    It provides a minimal wrapper on top of the existing python-bindings to conform with
    the way mumott parametrizes the geometry.

    Parameters
    ----------
    geometry : Geometry
        An instance of :class:`Geometry <mumott.Geometry>` containing the
        necessary vectors to compute forwared and adjoint projections.
    """

    def __init__(self,
                 geometry: Geometry):
        super().__init__(geometry)
        self._has_astra_objects = False
        self._update(force_update=True)

    def _geometry_to_astra_vec(self):
        """
        Transform geometric information as represented in mumott to the way
        it is represented in ASTRA.
        """
        # rename things
        geom = self._geometry
        number_of_projections = len(geom.rotations)
        p_0 = geom.p_direction_0
        j_0 = geom.j_direction_0
        k_0 = geom.k_direction_0

        # initialize vec_array
        self._astra_vec = np.zeros((number_of_projections, 12))

        # loop over all projections. (Could be sped up by vectorization.)
        for i in range(number_of_projections):
            R = geom._rotations[i].T
            p = R @ p_0
            j = R @ j_0
            k = R @ k_0
            shift_in_xyz = j * geom.j_offsets[i] + k * geom.k_offsets[i]

            # make ASTRA-style output
            self._astra_vec[i, :] = (*p, *shift_in_xyz, *j, *k)

    def _set_up_astra_objects(self):
        """
        Initialize ASTRA object (allocate arrays on the graphics card)
        """
        # (in astra x means index number 2!)
        self.proj_geom = astra.create_proj_geom('parallel3d_vec',
                                                self._geometry.projection_shape[1],
                                                self._geometry.projection_shape[0],
                                                self._astra_vec)

        # Astra syntax is (y, x, z)
        # See https://www.astra-toolbox.com/docs/geom3d.html
        self.vol_geom = astra.create_vol_geom(self._geometry.volume_shape[1],
                                              self._geometry.volume_shape[0],
                                              self._geometry.volume_shape[2])

        # Allocate arrays on the GPU with zeros
        self.astra_volume_id = astra.data3d.create('-vol', self.vol_geom)
        self.astra_sino_id = astra.data3d.create('-sino', self.proj_geom)

        # Set up the projection and back-projection algorithms
        config = dict(type='FP3D_CUDA',
                      ProjectionDataId=self.astra_sino_id,
                      VolumeDataId=self.astra_volume_id)
        self.proj_algorithm_id = astra.algorithm.create(config)

        config = dict(type='BP3D_CUDA',
                      ProjectionDataId=self.astra_sino_id,
                      ReconstructionDataId=self.astra_volume_id)
        self.bp_algorithm_id = astra.algorithm.create(config)

    def forward(self,
                field: NDArray,
                indices: NDArray[int] = None) -> NDArray:
        """Compute the forward projection of a tensor field.

        Parameters
        ----------
        field
            An array containing coefficients in its fourth dimension,
            which are to be projected into two dimensions. The first three
            dimensions should match the :attr:`volume_shape` of the sample.
        indices
            Must be ``None`` since the computation of individual projections
            is not supported for this projector.

        Returns
        -------
            An array with four dimensions ``(I, J, K, L)``, where
            the first dimension corresponds to projection index.
            The second and third dimension contain the pixels in the
            ``J`` and ``K`` dimension respectively, whereas
            the last dimension is the coefficient dimension, matching
            ``field[-1]``.
        """
        if not np.allclose(field.shape[:-1], self._geometry.volume_shape):
            raise ValueError(f'The shape of the input field ({field.shape}) does not match the'
                             f' volume shape expected by the projector ({self._geometry.volume_shape})')
        self._update()
        if indices is None:
            return self._forward_stack(field)
        else:
            raise NotImplementedError('The SAXSProjectorAstra object does not support the'
                                      ' computation of individual projections.')

    def _forward_stack(self,
                       volume: NDArray) -> NDArray:
        """Internal method for forward projecting an entire stack.

        Parameters
        ----------
        field
            The field to be projected.

        Returns
        -------
            The resulting projections.
        """
        if len(volume.shape) == 3:
            # Scalar projection of a 3D volume

            # move data to GPU
            astra.data3d.store(self.astra_volume_id, volume.transpose((2, 1, 0)))
            astra.data3d.store(self.astra_sino_id, 0)

            # carry out projection
            astra.algorithm.run(self.proj_algorithm_id)

            # move data back to system RAM and return
            return astra.data3d.get(self.astra_sino_id).transpose((1, 2, 0))

        elif len(volume.shape) == 4:
            # initialize output array
            number_of_channels = volume.shape[3]
            output_array = np.zeros((len(self._geometry.rotations),
                                     *self._geometry.projection_shape,
                                     number_of_channels))

            # loop over channels and carry out a scalar projection
            for channel_number in range(volume.shape[3]):
                output_array[..., channel_number] = self._forward_stack(volume[..., channel_number])

            return output_array

    def adjoint(self,
                projections: NDArray,
                indices: NDArray[int] = None) -> NDArray:
        """Compute the adjoint of a set of projections according to the system geometry.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from e.g. the residual of measured data and forward projections.
            The first dimension should match :attr:`indices` in size, and the
            second and third dimensions should match the system projection geometry.
        indices
            A one-dimensional array containing one or more indices
            indicating from which projections the adjoint is to be computed.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial and the last dimension runs over
            coefficients.
        """
        if not np.allclose(projections.shape[-3:-1], self._geometry.projection_shape):
            raise ValueError(f'The shape of the projections ({projections.shap}) does not match the'
                             f' volume shape expected by the projector ({self._geometry.projection_shape})')
        self._update()

        if indices is None or indices.size == len(self._geometry):
            return self._adjoint_stack(projections)
        else:
            raise NotImplementedError('The SAXSProjectorAstra object does not support the'
                                      ' computation of individual projections.')

        return self._adjoint_subset(projections, indices)

    def _adjoint_stack(self,
                       sinograms: NDArray) -> NDArray:
        """Internal method for computing the adjoint of a whole stack of projections.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from e.g. the residual of measured data and forward projections.
            The first dimension should run over all the projection directions
            in the system geometry.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial, and the last dimension runs over
            coefficients.
        """

        if len(sinograms.shape) == 3:
            # Scalar back-projection

            # copy data to GPU RAM
            astra.data3d.store(self.astra_volume_id, 0)
            astra.data3d.store(self.astra_sino_id, sinograms.transpose((2, 0, 1)))

            # carry out back-projection
            astra.algorithm.run(self.bp_algorithm_id)

            # move data back to system RAM and return
            return astra.data3d.get(self.astra_volume_id).transpose((2, 1, 0))

        elif len(sinograms.shape) == 4:
            # initialize output array
            number_of_channels = sinograms.shape[3]
            output_array = np.zeros((*self._geometry.volume_shape, number_of_channels))

            # loop over channels and carry out scalar projection
            for channel_number in range(sinograms.shape[3]):
                output_array[..., channel_number] = self._adjoint_stack(
                    sinograms[..., channel_number])

            return output_array

    def _update(self,
                force_update: bool = False) -> None:
        super()._update(force_update)  # This quits the function if not dirty (hidden in .super)

        self._geometry_to_astra_vec()
        if self._has_astra_objects:
            astra.astra.delete([self.astra_volume_id,
                                self.astra_sino_id,
                                self.proj_algorithm_id,
                                self.bp_algorithm_id])

        self._set_up_astra_objects()
        self._has_astra_objects = True
        self._geometry_hash = hash(self._geometry)

    def __hash__(self) -> int:
        to_hash = [self._geometry_hash, hash(self._geometry)]
        return int(list_to_hash(to_hash), 16)

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += [self.__class__.__name__.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, edgeitems=2, precision=5, linewidth=60):
            s += ['{:18} : {}'.format('is_dirty', self.is_dirty)]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += [f'<h3>{self.__class__.__name__}</h3>']
        s += ['<table border="1" class="dataframe">']
        s += [
            '<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>'
        ]
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">is_dirty</td>']
            s += [f'<td>1</td><td>{self.is_dirty}</td></tr>']
            s += ['<tr><td style="text-align: left;">hash</td>']
            s += [
                f'<td>{len(hex(hash(self)))}</td><td>{hex(hash(self))[2:8]}</td></tr>'
            ]
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)



""" Numba version of John transform. """

from numba import njit, int32, float64, void, prange, optional
from numpy.typing import NDArray
from typing import Optional
import numba
import numpy as np

numba.config.THREADING_LAYER = 'omp'


@njit(int32(float64[:, ::1], float64[::1]), fastmath=True, cache=True, nogil=True)
def _check_bounds(position: NDArray[float64],
                  bounds: NDArray[float64]) -> int32:
    """ Checks whether the given positions are inside a rectangular volume
    defined by the two corners at ``(0, 0, 0)`` and ``bounds``.

    Parameters
    ----------
    position
        ``position[i, :]`` gives the ``(x, y, z)`` coordinates
        of each point in space to be checked.
    bounds
        Contains the ``(x, y, z)`` coordinates for the three planes
        that define the rectangular volume.

    Returns
    -------
    out
        ``True`` if the positions are all inside the volume,
        ``False`` if at least one position is outside the volume.
    """
    out = int32(1)
    for h in range(position.shape[0]):
        out &= (int32(position[h, 0]) >= 0) & \
            (int32(position[h, 0]) < bounds[0]) & \
            (int32(position[h, 1]) >= 0) & \
            (int32(position[h, 1]) < bounds[1]) & \
            (int32(position[h, 2]) >= 0) & \
            (int32(position[h, 2]) < bounds[2])
    return out


@njit(void(float64[:, ::1], int32, float64[::1]), fastmath=True, cache=True, nogil=True)
def _update_position(position: NDArray[float64],
                     steps: int32,
                     projection_vector: NDArray[float64]) -> None:
    """ Updates the position for a set of rays with
    the specified number of additions of the given projection vector.

    Parameters
    ----------
    position
        ``position[i, :]`` contains the starting ``(x, y, z)`` coordinates of
        ray ``i``. This is modified in-place to update the position.
    steps
        Number of additions of ``projection_vector`` that are performed.
    projection_vector
        Determines the size and direction of each step.
    """
    for i in range(position.shape[0]):
        position[i, 0] += projection_vector[0] * steps
        position[i, 1] += projection_vector[1] * steps
        position[i, 2] += projection_vector[2] * steps


@njit(int32[:, ::1](float64[:, ::1], float64[::1], float64[::1], int32),
      fastmath=True, cache=True, nogil=True)
def _get_indices(position: NDArray[float64],
                 projection_vector: NDArray[float64],
                 volume_indexer: NDArray[float64],
                 steps: int32) -> NDArray[int32]:
    """ Computes all the (flattened) indices sampled by each ray of
    integration. Does not do bounds checks, so use with caution.

    Parameters
    ----------
    position
        ``position[i, :]`` contains the starting ``(x, y, z)`` coordinates of
        ray ``i``. No checks are performed to ensure that the position
        remains inside the volume, so use with caution. The provided
        ``position`` array is not modified by this function.
    projection_vector
        Determines the size and direction of each integration step.
    volume_indexer
        Contains the three-dimensional ``(x, y, z)`` shape of the volume.
    steps
        Number of steps needed for each ray of integration to reach its
        end point. No checks are performed to ensure that the number of
        steps do not result in stepping outside the volume, so use with caution.

    Returns
    -------
    out_vector
        Contains the flattened indices, such that out_vector[i, j]
        is the index sampled by ray ``j`` after ``i```steps.
    """
    out_vector = np.empty((steps, position.shape[0]), dtype=np.int32)
    for i in range(steps):
        for j in range(position.shape[0]):
            out_vector[i, j] = int32(int32((position[j, 0] + i * projection_vector[0])) * volume_indexer[0] +
                                     int32((position[j, 1] + i * projection_vector[1])) * volume_indexer[1] +
                                     int32((position[j, 2] + i * projection_vector[2])) * volume_indexer[2])
    return out_vector


@njit(int32(int32[:, ::1], float64[::1], int32[::1], float64[::1]),
      fastmath=True, cache=True, nogil=True)
def _reduce_indices(indices: NDArray[int32],
                    sampling_kernel: NDArray[float64],
                    reduced_indices: NDArray[int32],
                    weights: NDArray[float64]) -> NDArray[int32]:
    """ Reduces repeated indices by accumulating their weights,
    in order to reduce the number of reads or writes to the volume.

    Parameters
    ----------
    indices
        Array of flat indices of the volume, with
        ``indices[i, j]`` referring to the index of the volume samped by the ray of
        integration with weight ``sampling_kernel[j]`` after ``i`` steps.
    sampling_kernel
        Array of weights for each ray of integration.
    reduced_indices
        Output variable with the reduced indices. Must be large enough to
        store all indices after reduction.
    weights
        Output variable with ``weight[i]`` containing the accumulated
        weight of ``reduced_indices[i]``. Must be large enough to store
        all weights after reduction.

    Returns
    -------
    counter
        The number of valid entries of ``reduced_indices`` and
        ``weights``, such that ``reduced_indices[counter - 1]`` contains
        the last sampled index.
    """
    current_index = np.int32(-1)
    counter = 0
    ii = 0
    for i in range(indices.shape[0]):
        for j in range(indices.shape[1]):
            if current_index == indices[i, j]:
                weights[counter - 1] += sampling_kernel[j]
            else:
                if reduced_indices[ii] == indices[i, j]:
                    weights[ii] += sampling_kernel[j]
                else:
                    for ii in range(counter - 2, counter - sampling_kernel.size - 1, -1):
                        if reduced_indices[ii] == indices[i, j]:
                            weights[ii] += sampling_kernel[j]
                            break
                    else:
                        current_index = indices[i, j]
                        reduced_indices[counter] = indices[i, j]
                        weights[counter] = sampling_kernel[j]
                        counter += 1
    return counter


@njit(void(int32, int32, float64[:, ::1], float64[::1],
           float64[::1], float64[::1], float64[::1],
           float64, float64[::1], float64[:, ::1]), fastmath=True, cache=True, nogil=True)
def _initialize_position(j: int32, k: int32, offsets: NDArray[float64],
                         projection_bounds: NDArray[float64],
                         unit_vector_p: NDArray[float64],
                         unit_vector_j: NDArray[float64],
                         unit_vector_k: NDArray[float64],
                         scalar_bound: NDArray[float64],
                         volume_shift: NDArray[float64],
                         position: NDArray[float64]) -> None:
    """ Initializes the position of a ray of integration
    by accounting for the volume geometry, projection geometry,
    projection alignment offsets, and sampling kernel offsets.

    Parameters
    ----------
    j
        The first index of the pixel in the projection.
    k
        The second index of the pixel in the projection.
    offsets
        Offsets for each ray of the sampling kernel.
    projection_bounds
        The spatial position of index ``(0, 0)`` of the
        projection after aligning the point on the projection
        that includes the center of the volume with the origin,
        in the directions of ``unit_vector_j`` and ``unit_vector_k``.
    scalar_bound
        The offset in the direction of ``unit_vector_p``, from
        the plane that intersects the center of the volume.
    volume_shift
        The position of the center of the volume.
    position
        Output variable containing positions for each ray.
        Modified in-place.

    Notes
    -----
    It is possible to use an alignment which aligns some other
    point of the volume and the projection. In that case,
    ``projection_bounds``, ``scalar_bound`` and ``volume_shift``
    would all need to be modified to refer to this point of alignment
    rather than the center.
    """
    shift_j = j + 0.5 + projection_bounds[0]
    shift_k = k + 0.5 + projection_bounds[1]
    for i in range(position.shape[0]):
        position[i, 0] = (shift_j + offsets[0, i]) * unit_vector_j[0] + \
            (shift_k + offsets[1, i]) * unit_vector_k[0] + \
            scalar_bound * unit_vector_p[0] + volume_shift[0]
        position[i, 1] = (shift_j + offsets[0, i]) * unit_vector_j[1] + \
            (shift_k + offsets[1, i]) * unit_vector_k[1] + \
            scalar_bound * unit_vector_p[1] + volume_shift[1]
        position[i, 2] = (shift_j + offsets[0, i]) * unit_vector_j[2] + \
            (shift_k + offsets[1, i]) * unit_vector_k[2] + \
            scalar_bound * unit_vector_p[2] + volume_shift[2]


@njit(int32(float64[:, ::1], float64[::1], float64[::1]), fastmath=True, cache=True, nogil=True)
def _get_start_and_end_points(position: NDArray[float64],
                              projection_vector: NDArray[float64],
                              volume_bounds: NDArray[float64]) -> NDArray[int32]:
    """ Places the positions of each ray inside the volume bounds,
    and returns the number of steps needed to reach the edge
    of the volume.

    Parameters
    ----------
    position
        A set of ``(x, y, z)`` positions at which each parallel ray begins.
        If inside the volume, the region opposite to ``projection_vector``
        will not be integrated over. If outside the volume, all positions
        will be shifted to the inside of the volume.
    projection_vector
        The direction in which each ray is to travel, with the magnitude
        of the vector giving the step size.
    volume_bounds
        The ``(x, y, z)`` shape of thre three-dimensional volume
        that the rays are to travel through.

    Returns
    -------
    end_steps
        The number of steps needed before the first ray reaches
        the end of the volume, such that
        ``position[i, :] + end_steps * projection_vector``
        will be inside the volume for all ``i``, and
        ``position[i, :] + (end_steps + 1) * projection_vector``
        will be outside the volume for at least one ``i``.
    """
    adj_vector = np.empty(3)
    for i in range(3):
        if projection_vector[i] == 0:
            adj_vector[i] = np.finfo(float64).max
        else:
            adj_vector[i] = float64(1) / projection_vector[i]
    upper_length_bound = volume_bounds[0] * volume_bounds[1] * volume_bounds[2]
    bounds = np.empty(6, dtype=np.float64)
    for i in range(3):
        bounds[i] = volume_bounds[i] * (projection_vector[i] < 0)
        bounds[i + 3] = volume_bounds[i] * (projection_vector[i] >= 0)
    step_diff = float64(0.)
    temp_diff = float64(0.)
    for i in range(position.shape[0]):
        for j in range(3):
            temp_diff = (bounds[j] - position[i, j]) * adj_vector[j]
            step_diff = max(step_diff, temp_diff)
    start_steps = int32(step_diff + 1)
    if start_steps > upper_length_bound:
        return int32(0)
    _update_position(position, start_steps, projection_vector)
    bounds_okay = _check_bounds(position, volume_bounds)
    if not bounds_okay:
        return int32(0)
    step_diff = float64(volume_bounds[0] * volume_bounds[1] * volume_bounds[2])
    temp_diff = float64(0.)
    for i in range(position.shape[0]):
        for j in range(3):
            temp_diff = (bounds[j + 3] - position[i, j]) * adj_vector[j]
            step_diff = min(temp_diff, step_diff)
    end_steps = int32(step_diff)
    if end_steps <= 0:
        return int32(0)
    _update_position(position, end_steps, projection_vector)
    bounds_okay &= _check_bounds(position, volume_bounds)
    if bounds_okay:
        _update_position(position, -end_steps, projection_vector)
        return end_steps
    else:
        return int32(0)


@njit(void(float64[:, ::1], float64[::1], float64[:, ::1],
           float64[::1], float64[::1], int32, float64[::1]), fastmath=True, cache=True, nogil=True)
def _integrate(position: NDArray[float64], sampling_kernel: NDArray[float64],
               volume: NDArray[float64], volume_indexer: NDArray[float64],
               projection_vector: NDArray[float64], total_steps: int32,
               pixel: NDArray[float64]) -> None:
    """ Performs the integration necessary to compute the
    the John transform.

    Parameters
    ----------
    position
        ``position[i, :]`` is the position of the ray with weight ``sampling_kernel[i]``
        in ``(x, y, z)`` coordinates. No checks are performed to verify
        that the position is inside the volume, so proceed with caution.
    sampling_kernel
        Contains the weights for each ray of integration.
    volume
        Two-dimensional volume. The first index
        corresponds to the internal coordinates of the volume
        determined by ``volume_indexer``, and the final index
        stores multichannel data identically to ``pixel``.
    volume_indexer
        ``volume_indexer[i]`` gives the ``[x, y, z]`` coordinates
        of ``volume[i, :]``.
    projection_vector
        The direction of projection in Cartesian coordinates,
        scaled according to integration step size.
    total_steps
        Total number of additions of ``projection_vector``
        needed to reach the end of the
        volume from the starting position, for the ray which
        would need the least number of steps to reach it.
        No checks are performed to see if the rays remain
        inside the volume, so proceed with caution.
    pixel
        Output variable where the projection of each
        channel of ``volume`` is stored.
    """
    indices = _get_indices(position, projection_vector, volume_indexer, total_steps)
    reduced_indices = np.empty(indices.size, dtype=np.int32)
    weights = np.empty(indices.size, dtype=np.float64)
    counter = _reduce_indices(indices, sampling_kernel, reduced_indices, weights)
    for i in range(counter):
        for j in range(pixel.size):
            pixel[j] += volume[reduced_indices[i], j] * weights[i]


@njit(void(float64[:, ::1], float64[::1], float64[::1], float64[:, ::1],
           float64[::1], float64[::1], int32), fastmath=True, cache=True, nogil=True)
def _back_integrate(position: NDArray[float64], distributor: NDArray[float64],
                    sampling_kernel: NDArray[float64], volume: NDArray[float64],
                    volume_indexer: NDArray[float64], projection_vector: NDArray[float64],
                    total_steps: int32) -> None:
    r""" Performs the integration necessary to compute the
    adjoint of the John transform.

    Parameters
    ----------
    position
        ``position[i, :]`` is the position of the ray with weight ``sampling_kernel[i]``
        in ``(x, y, z)`` coordinates. No checks are performed to verify
        that the position is inside the volume, so proceed with caution.
    distributor
        Contains the values to be distributed to each channel of the volume
        along the ray of integration.
    sampling_kernel
        Contains the weights for each ray of integration.
    volume
        Two-dimensional volume. The first index
        corresponds to the internal coordinates of the volume
        determined by ``volume_indexer``, and the final index
        stores multichannel data identically to ``distributor``.
    volume_indexer
        ``volume_indexer[i]`` gives the ``[x, y, z]`` coordinates
        of ``volume[i, :]``.
    projection_vector
        The direction of projection in Cartesian coordinates,
        scaled according to integration step size.
    total_steps
        Total number of additions of ``projection_vector``
        needed to reach the end of the
        volume from the starting position, for the ray which
        would need the least number of steps to reach it.
        No checks are performed to see if the rays remain
        inside the volume, so proceed with caution.
    """
    indices = _get_indices(position, projection_vector, volume_indexer, total_steps)
    reduced_indices = np.empty(indices.size, dtype=np.int32)
    weights = np.empty(indices.size, dtype=np.float64)
    counter = _reduce_indices(indices, sampling_kernel, reduced_indices, weights)
    for i in range(counter):
        for j in range(distributor.size):
            volume[reduced_indices[i], j] += weights[i] * distributor[j]


@njit(void(float64[:, ::1],
           float64[::1],
           float64[::1], float64[::1],
           float64[::1], float64[::1], float64[::1],
           float64[:, ::1], float64[::1]), fastmath=True, cache=True, nogil=True)
def _project_for_one_pixel(shifted_positions: NDArray[float64],
                           projection_vector: NDArray[float64],
                           unit_vector_j: NDArray[float64], unit_vector_k: NDArray[float64],
                           sampling_kernel: NDArray[float64], volume_bounds: NDArray[float64],
                           pixel: NDArray[float64], volume: NDArray[float64],
                           volume_indexer: NDArray[float64]) -> None:
    """ Computes the projection into one pixel in the projection.

    Parameters
    ----------
    shifted_positions
        ``shifted_positions[i, :]`` initial ``(x, y, z)`` positions
        for the projection ray with weight ``sampling_kernel[i]``.
    projection_vector
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    volume_bounds
        Contains the non-flattened shape of dimension 0 of ``volume``.
    sampling_kernel
        A kernel for upsampling or convolution of the projection,
        scaled by the size of each integration step. For a single ray,
        this array has only one element containing the step size.
    pixel
        The pixel into which the projection is calculated, with
        multichannel indices.
    volume
        Two-dimensional volume variable. The first index
        corresponds to the internal coordinates of the volume
        determined by ``volume_indexer``, and the final index
        stores multichannel data identically to ``pixel``.
    volume_indexer
        ``volume_indexer[i]`` gives the ``[x, y, z]`` coordinates
        of ``volume[i, :]``.
    """
    steps = _get_start_and_end_points(shifted_positions, projection_vector, volume_bounds)
    if steps == 0:
        return
    _integrate(shifted_positions, sampling_kernel,
               volume, volume_indexer,
               projection_vector, steps, pixel)


@njit(void(float64[:, ::1], float64[::1], float64[::1],
           float64[::1], float64[::1],
           float64[::1], float64[::1],
           float64[:, ::1], float64[::1]), fastmath=True, cache=True, nogil=True)
def _back_project_for_one_pixel(shifted_positions: NDArray[float64], distributor: NDArray[float64],
                                sampling_kernel: NDArray[float64],
                                projection_vector: NDArray[float64],
                                unit_vector_j: NDArray[float64], unit_vector_k: NDArray[float64],
                                volume_bounds: NDArray[float64],
                                volume: NDArray[float64], volume_indexer: NDArray[float64]) -> None:
    """ Computes the adjoint from one pixel in the projection.

    Parameters
    ----------
    shifted_positions
        ``shifted_positions[i, :]`` initial ``(x, y, z)`` positions
        for the projection ray with weight ``sampling_kernel[i]``.
    distributor
        Contains the contribution to the adjoint for the pixel
        in question, for each multichannel index.
    sampling_kernel
        A kernel for upsampling or convolution of the projection,
        scaled by the size of each integration step. For a single ray,
        this array has only one element equal to the step size.
    projection_vector
        The direction of projection in Cartesian coordinates,
        scaled by the size of each integration step.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    volume_bounds
        Contains the non-flattened shape of dimension 0 of ``volume``.
    volume
        Two-dimensional volume variable. The first index
        corresponds to the internal coordinates of the volume
        determined by ``volume_indexer``, and the final index
        stores multichannel data identically to ``distributor``.
    volume_indexer
        ``volume_indexer[i]`` gives the ``[x, y, z]`` coordinates
        of ``volume[i, :]``.
    """
    steps = _get_start_and_end_points(shifted_positions, projection_vector, volume_bounds)
    if steps == 0:
        return
    _back_integrate(shifted_positions, distributor, sampling_kernel,
                    volume, volume_indexer,
                    projection_vector, steps)


@njit(void(float64[:, :, ::1], float64[::1], float64[:, ::1],
           float64[::1], float64[::1], float64[::1], float64[::1], float64[::1],
           float64[::1], float64[:, ::1], float64, float64),
      parallel=True, fastmath=True, cache=True, nogil=True)
def _john_transform(projection: NDArray[float64], projection_offsets: NDArray[float64],
                    volume: NDArray[float64], volume_bounds: NDArray[float64],
                    volume_indexer: NDArray[float64], unit_vector_p: NDArray[float64],
                    unit_vector_j: NDArray[float64], unit_vector_k: NDArray[float64],
                    sampling_kernel: NDArray[float64], kernel_offsets: NDArray[float64],
                    step_size: NDArray[float64], scalar_bound: NDArray[float64]) -> None:
    """ Internal method for computing the John transform
    with CPU multithreading.

    Parameters
    ----------
    projection
        A 3-dimensional numpy array for which the adjoint is calculated.
        The last index contains the
        data channels for multi-channel data. For scalar data, use
        ``1`` as the size of the last dimension.
    projection_offsets
        The j, k offsets that align the projections.
    volume
        Two-dimensional volume variable. The first index
        corresponds to the internal coordinates of the volume
        determined by ``volume_indexer``, and the final index
        stores multichannel data identically to the last
        index of ``projection``.
    volume_bounds
        Contains the non-flattened shape of dimension 0 of ``volume``.
    volume_indexer
        ``volume_indexer[i]`` gives the ``[x, y, z]`` coordinates
        of ``volume[i, :]``.
    unit_vector_p
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    sampling_kernel
        A kernel for upsampling or convolution of the projection.
        Should be normalized so that it sums to unity.
    kernel_offsets
        Offsets for each entry of the kernel.
    step_size
        The size of each step for the quadrature of the integral.
    scalar_bound
        The offset from the center of the volume at which the
        projection begins, in the direction of ``unit_vector_p``.
    """
    projection_vector = unit_vector_p * step_size
    kernel = sampling_kernel * step_size
    volume_shift = float64(0.5) * volume_bounds
    projection_shift = (projection_offsets - np.array(projection.shape[:-1]) * 0.5).astype(np.float64)
    for j in prange(projection.shape[0]):
        shifted_positions = np.empty((sampling_kernel.size, 3), dtype=np.float64)
        for k in range(projection.shape[1]):
            _initialize_position(j, k, kernel_offsets, projection_shift,
                                 unit_vector_p, unit_vector_j, unit_vector_k,
                                 scalar_bound, volume_shift,
                                 shifted_positions)
            _project_for_one_pixel(shifted_positions,
                                   projection_vector,
                                   unit_vector_j, unit_vector_k,
                                   kernel, volume_bounds,
                                   projection[j, k, ...], volume, volume_indexer)


@njit(void(float64[:, :, ::1], float64[::1], float64[:, :, ::1],
           float64[::1], float64[::1], float64[::1], float64[::1], float64[::1],
           float64[::1], float64[:, ::1], float64, float64),
      parallel=True, fastmath=True, cache=True, nogil=True)
def _john_transform_adjoint(projection: NDArray[float64], projection_offsets: NDArray[float64],
                            volume: NDArray[float64], volume_bounds: NDArray[float64],
                            volume_indexer: NDArray[float64], unit_vector_p: NDArray[float64],
                            unit_vector_j: NDArray[float64], unit_vector_k: NDArray[float64],
                            sampling_kernel: NDArray[float64], kernel_offsets: NDArray[float64],
                            step_size: float64, scalar_bound: float64) -> None:
    """ Internal method for computing the adjoint of the John transform
    with CPU multithreading. All arrays must be row major and contiguous,
    also called C-contiguous.

    Parameters
    ----------
    projection
        A 3-dimensional numpy array for which the adjoint is calculated.
        The last index contains the
        data channels for multi-channel data. For scalar data, use
        ``1`` as the size of the last dimension.
    projection_offsets
        The j, k offsets that align the projections.
    volume
        Three-dimensional volume variable. The first index
        must equal ``numba.get_num_threads``, the second index
        correspond to the internal coordinates of the volume
        determined by ``volume_indexer``, and the final index
        stores multichannel data identically to the last
        index of ``projection``.
    volume_bounds
        Contains the non-flattened shape of dimension 1 of ``volume``.
    volume_indexer
        ``volume_indexer[i]`` gives the ``[x, y, z]`` coordinates
        of ``volume[:, i, :]``.
    unit_vector_p
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    sampling_kernel
        A kernel for upsampling or convolution of the projection.
        Should be normalized so that it sums to unity.
    kernel_offsets
        Offsets for each entry of the kernel.
    step_size
        The size of each step for the quadrature of the integral.
    scalar_bound
        The offset from the center of the volume at which the
        projection begins, in the direction of ``unit_vector_p``.
    """
    projection_vector = unit_vector_p * step_size
    kernel = sampling_kernel * step_size
    volume_shift = float64(0.5) * volume_bounds
    projection_shift = (projection_offsets - np.array(projection.shape[:-1]) * 0.5).astype(np.float64)
    for j in prange(projection.shape[0]):
        thread_volume = volume[numba.get_thread_id()]
        shifted_positions = np.empty((sampling_kernel.size, 3), dtype=np.float64)
        for k in range(projection.shape[1]):
            _initialize_position(j, k, kernel_offsets, projection_shift,
                                 unit_vector_p, unit_vector_j, unit_vector_k,
                                 scalar_bound, volume_shift,
                                 shifted_positions)
            distributor = projection[j, k, ...]
            _back_project_for_one_pixel(shifted_positions, distributor, kernel,
                                        projection_vector,
                                        unit_vector_j, unit_vector_k,
                                        volume_bounds, thread_volume, volume_indexer)


@njit(void(float64[:, :, ::1], float64[:, :, :, ::1],
           float64[::1], float64[::1], float64[::1],
           float64, float64[::1],
           optional(float64[::1]), optional(float64[:, ::1])),
      fastmath=True, cache=True, nogil=True)
def john_transform(projection: NDArray[float64],
                   volume: NDArray[float64],
                   unit_vector_p: NDArray[float64],
                   unit_vector_j: NDArray[float64],
                   unit_vector_k: NDArray[float64],
                   step_size: float64,
                   projection_offsets: NDArray[float64],
                   sampling_kernel: Optional[NDArray[float64]],
                   kernel_offsets: Optional[NDArray[float64]]) -> None:
    r""" Frontend for performing the John transform with parallel
    CPU computing. The number of threads is regulated with
    ``numba.set_num_threads()``. All arrays must be row-major and contiguous,
    also called C-contiguous.

    Parameters
    ----------
    projection
        A 3-dimensional numpy array where the projection is stored.
        The last index contains the
        data channels for multi-channel data. For scalar data, use
        ``1`` as the size of the last dimension.
    volume
        The volume to be projected, with 4 dimensions. The last index
        is the same as for ``projection``.
    unit_vector_p
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    step_size
        The size of each step for the quadrature of the integral.
    projection_offsets
        The j, k offsets that align the projections.
    sampling_kernel
        A kernel for uppsampling or convolution of the projection.
        Should be normalized so that it sums to unity.
    kernel_offsets
        Offsets for each entry of the kernel.

    Notes
    -----
    The computation performed by this function may be written as

    .. math::

        p(J, K)_i = \sum_{s=0}^{N} \sum_{j=0}^{M} d \cdot w_j \cdot V_i(\lfloor \mathbf{r}_j + d s \cdot \mathbf{v} \rfloor)

    where :math:`p(J, K)_i` is ``projection[J, K, i]``, :math:`d` is `step_size`,
    :math:`N` is the total number of steps, :math:`M` is ``sampling_kernel.size``,
    :math:`w_j` is ``sampling_kernel[j]``,  :math:`V_i` is ``volume[..., i]``,
    and :math:`\mathbf{v}` is ``unit_vector_p``.
    \mathbf{r}_j is the starting position for the ray weighted by ``sampling_kernel[j]``,
    and is given by

    .. math::

        \mathbf{r}_j(J, K) = (J + 0.5 + o_J - 0.5 \cdot J_\text{max}) \cdot \mathbf{u} + \\
                (K + 0.5 + o_K - 0.5 \cdot K_\text{max}) \cdot \mathbf{w} + \\
                (\Delta_p) \cdot \mathbf{v} + 0.5 \cdot \mathbf{r}_\text{max} \\

    where :math:`o_J` is the sum of ``projection_offsets[0]`` and ``kernel_offsets[0, j]``,
    :math:`o_K` is the sum of ``projection_offsets[1]`` and ``kernel_offsets[1, j]``,
    :math:`\mathbf{u}` is ``unit_vector_j``, :math:`mathbf{w}` is ``unit_vector_k``,
    :math:`J_\text{max}` and :math`K_\text{max}` are ``projection.shape[0]`` and
    ``projection.shape[1]`` respectively, and :math:`\mathbf{r}_\text{max}` is
    ``volume.shape[:3]``. :math:`\Delta_p` is an additional offset that places the
    starting position at the edge of the volume.
    """ # noqa
    scalar_bound = -((1. + np.sqrt(np.sum(np.power(float64(volume.shape[:-1]), 2.)))))
    volume_bounds = np.array(volume.shape[:-1]).astype(np.float64)
    volume_indexer = np.flip(np.cumprod(np.array((1,) + volume.shape[-2:0:-1]))).astype(np.float64)
    volume = volume.reshape(-1, volume.shape[-1])
    if sampling_kernel is None or kernel_offsets is None:
        kernel = np.array((1,), dtype=float64)
        offsets = np.array(((0, 0),), dtype=float64)
    else:
        kernel = sampling_kernel
        offsets = kernel_offsets
    _john_transform(projection, projection_offsets,
                    volume, volume_bounds, volume_indexer,
                    unit_vector_p,
                    unit_vector_j, unit_vector_k,
                    kernel, offsets,
                    step_size, scalar_bound)


@njit(void(float64[:, :, ::1], float64[:, :, :, :, ::1],
           float64[::1], float64[::1], float64[::1],
           float64, float64[::1],
           optional(float64[::1]), optional(float64[:, ::1])),
      fastmath=True, cache=True, nogil=True)
def john_transform_adjoint(projection: NDArray[float64],
                           volume: NDArray[float64],
                           unit_vector_p: NDArray[float64],
                           unit_vector_j: NDArray[float64],
                           unit_vector_k: NDArray[float64],
                           step_size: NDArray[float64],
                           projection_offsets: NDArray[float64],
                           sampling_kernel: Optional[NDArray[float64]],
                           kernel_offsets: Optional[NDArray[float64]]) -> None:
    r""" Frontend for performing the John transform adjoint (back-projection)
    with parallel CPU computation. The number of threads is regulated by
    ``numba.set_num_threads``.

    Parameters
    ----------
    projection
        A 3-dimensional numpy array for which the adjoint is calculated.
        The last index contains the
        data channels for multi-channel data. For scalar data, use
        ``1`` as the size of the last dimension.
    volume
        Output variable modified in-place. The volume where the adjoint is stored,
        with 5 dimensions. The last index is the same as the last index in
        ``projection``. The first index is for each thread of
        multi-threaded computation, so ``volume.shape[0]`` must equal
        ``numba.get_num_threads``.
        After computing, it is necessary to reduce over the first
        index to obtain the total result.
    unit_vector_p
        The direction of projection in Cartesian coordinates.
    unit_vector_j
        One of the directions for the pixels of ``projection``.
    unit_vector_k
        The other direction for the pixels of ``projection``.
    step_size
        The size of each step for the quadrature of the integral.
    projection_offsets
        The j, k offsets that align the projections.
    sampling_kernel
        A kernel for upsampling or convolution of the projection.
        Should be normalized so that it sums to unity.
    kernel_offsets
        Offsets for each entry of the kernel.

    Notes
    -----
    The computation performed by this function may be written as

    .. math::

        V_i(\mathbf{x}) = \sum_{s=0}^{N} \sum_{j=0}^{M} d \cdot w_j \cdot p(J, K)_i \delta_{\lfloor\mathbf{r}_j + d s \cdot \mathbf{v} \rfloor}^{\mathbf{x}}

    where :math:`d_i` is ``distributor[i]``, :math:`N` is the total number of steps,
    :math:`d` is ``step_size``. :math:`M` is ``sampling_kernel.size``, :math:`w_j` is ``sampling_kernel[j]``,
    :math:`V_i` is ``volume[:, i]``,
    and :math:`\mathbf{v}` is ``projection_vector``. :math:`\mathbf{x}` is
    understood to run over every triplet of non-negative integers in the domain
    of :math:`V_i`. \mathbf{r}_j is the starting position for the ray weighted by ``sampling_kernel[j]``,
    and is given by

    .. math::

        \mathbf{r}_j(J, K) = (J + 0.5 + o_J - 0.5 \cdot J_\text{max}) \cdot \mathbf{u} + \\
                (K + 0.5 + o_K - 0.5 \cdot K_\text{max}) \cdot \mathbf{w} + \\
                (\Delta_p) \cdot \mathbf{v} + 0.5 \cdot \mathbf{r}_\text{max} \\

    where :math:`o_J` is the sum of ``projection_offsets[0]`` and ``kernel_offsets[0, j]``,
    :math:`o_K` is the sum of ``projection_offsets[1]`` and ``kernel_offsets[1, j]``,
    :math:`\mathbf{u}` is ``unit_vector_j``, :math:`mathbf{w}` is ``unit_vector_k``,
    :math:`J_\text{max}` and :math`K_\text{max}` are ``projection.shape[0]`` and
    ``projection.shape[1]`` respectively, and :math:`\mathbf{r}_\text{max}` is
    ``volume.shape[:3]``. :math:`\Delta_p` is an additional offset that places the
    starting position at the edge of the volume.
    """ # noqa
    scalar_bound = -(1. + np.sqrt(np.sum(np.power(float64(volume.shape[:-1]), 2.))))
    volume_bounds = np.array(volume.shape[1:-1]).astype(np.float64)
    volume_indexer = np.flip(np.cumprod(np.array((1,) + volume.shape[-2:1:-1]))).astype(np.float64)
    volume = volume.reshape(-1, int32(np.prod(volume_bounds)), volume.shape[-1])
    if sampling_kernel is None or kernel_offsets is None:
        kernel = np.array((1,), dtype=float64)
        offsets = np.array(((0, 0),), dtype=float64)
    else:
        kernel = sampling_kernel
        offsets = kernel_offsets
    _john_transform_adjoint(projection, projection_offsets,
                            volume, volume_bounds, volume_indexer,
                            unit_vector_p,
                            unit_vector_j, unit_vector_k,
                            kernel, offsets,
                            step_size, scalar_bound)



import pytest  # noqa

import numpy as np
from mumott import DataContainer

from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjectorNumba
from mumott.methods.functionals import SIGTT
from mumott.optimization.loss_functions import SquaredLoss

default_hash = 'b52d25'


def test_get_residual_norm():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    d = lf.get_residual_norm(get_gradient=False)
    assert np.isclose(d['residual_norm'], 7304.44)
    assert d['gradient'] is None

    d = lf.get_residual_norm(get_gradient=True)
    assert np.isclose(d['residual_norm'], 7304.44)
    assert np.allclose(d['gradient'][0, 0], [[-47.51], [-47.77], [-48.83], [-50.04]])

    dc.stack.weights *= 0.5
    lf = SquaredLoss(meth, use_weights=True)
    d = lf.get_residual_norm(get_gradient=False)
    assert np.isclose(d['residual_norm'], 3652.22)
    assert d['gradient'] is None

    d = lf.get_residual_norm(get_gradient=True)
    assert np.isclose(d['residual_norm'], 3652.22)
    assert np.allclose(d['gradient'][0, 0], [[-23.755], [-23.885], [-24.415], [-25.02]])

    lf.residual_norm_multiplier = 10
    d = lf.get_residual_norm(get_gradient=False)
    assert np.isclose(d['residual_norm'], 3652.22)

    lf.preconditioner = 0.5 * np.ones_like(meth.coefficients)
    lf.residual_norm_multiplier = 1
    d = lf.get_residual_norm(get_gradient=True)
    assert np.isclose(d['residual_norm'], 3652.22)
    assert np.allclose(d['gradient'][0, 0], np.array([[-23.755], [-23.885], [-24.415], [-25.02]]))

    lf.preconditioner = np.array((1, 2, 3))
    d = lf.get_residual_norm(get_gradient=True)


def test_get_loss():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    d = lf.get_loss(get_gradient=False)
    assert np.isclose(d['loss'], 7304.44)
    assert d['gradient'] is None

    d = lf.get_loss(get_gradient=True)
    assert np.isclose(d['loss'], 7304.44)
    assert np.allclose(d['gradient'][0, 0], [[-47.51], [-47.77], [-48.83], [-50.04]])

    dc.stack.weights *= 0.5
    lf = SquaredLoss(meth, use_weights=False)
    d = lf.get_loss(get_gradient=False)
    assert np.isclose(d['loss'], 7304.44)

    lf = SquaredLoss(meth, use_weights=True)
    d = lf.get_loss(get_gradient=False)
    assert np.isclose(d['loss'], 3652.22)

    d = lf.get_loss(get_gradient=False)
    lf.residual_norm_multiplier = 10
    d = lf.get_loss(get_gradient=False)
    assert np.isclose(d['loss'], 36522.24)

    lf.preconditioner = 0.5 * np.ones_like(meth.coefficients)
    lf.residual_norm_multiplier = 1
    d = lf.get_loss(get_gradient=True)
    assert np.isclose(d['loss'], 3652.22)
    assert np.allclose(d['gradient'][0, 0], 0.5 * np.array([[-23.755], [-23.885], [-24.415], [-25.02]]))

    lf.preconditioner = np.array((1, 2, 3))
    with pytest.raises(ValueError, match='first three dimensions'):
        d = lf.get_loss(get_gradient=True)


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    s = str(lf)
    assert default_hash in s


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    h = lf._repr_html_()
    assert default_hash in h


def test_hash():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth, use_weights=True)
    assert hex(hash(lf))[2:8] == '1d95cd'
    lf.use_weights = False
    assert hex(hash(lf))[2:8] == default_hash
    meth.coefficients += 1
    assert hex(hash(lf))[2:8] == '7dcb3b'



import pytest  # noqa:F401
from numba import cuda
import numpy as np

""" This test exists to check that the CI is compatible
with `numba.cuda`. """

test_array_1 = np.array([1, 2, 3]).astype(np.float32)
test_array_2 = np.array([0.5, -0.5, 1.5]).astype(np.float32)
result = np.array([1.5, 1.5, 4.5]).astype(np.float32)

device_array_1 = cuda.to_device(test_array_1)
device_array_2 = cuda.to_device(test_array_2)
out_device_array = cuda.device_array_like(device_array_1)

tpb = 3
bpg = test_array_1.size // tpb


@cuda.jit
def cuda_add(a, b, c):
    i = cuda.grid(1)
    if i < c.size:
        c[i] = a[i] + b[i]


@pytest.mark.parametrize('a,b,out,result', [(device_array_1, device_array_2,
                                             out_device_array, result)])
def test_cuda_add(a, b, out, result):
    cuda_add[tpb, bpg](a, b, out)
    assert np.allclose(out.copy_to_host(), result)



from abc import ABC, abstractmethod
from ... import ProbedCoordinates

from numpy.typing import NDArray


class BasisSet(ABC):

    """This is the base class from which specific basis sets are being derived.
    """

    def __init__(self,
                 probed_coordinates: ProbedCoordinates):
        pass

    @abstractmethod
    def forward(self,
                coefficients: NDArray,
                indices: NDArray = None) -> NDArray:
        pass

    @abstractmethod
    def gradient(self,
                 coefficients: NDArray,
                 indices: NDArray = None) -> NDArray:
        pass

    @abstractmethod
    def get_output(self,
                   coefficients: NDArray) -> dict:
        pass

    @abstractmethod
    def get_inner_product(self, u: NDArray, v: NDArray) -> NDArray:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __dict__(self) -> dict:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def _repr_html_(self) -> str:
        pass



import pytest # noqa
from mumott.data_handling import DataContainer
from mumott.output_handling import ProjectionViewer


def test():
    dc = DataContainer('tests/test_half_circle.h5')
    p = ProjectionViewer(dc)
    p.change_frame(0)



import pytest
import numpy as np

from mumott import ProbedCoordinates

vectors = [np.array((0.1, 1, 1), dtype=np.float64).reshape(1, 1, 1, 3),
           np.array((1, 2, 1), dtype=np.float64).reshape(1, 1, 1, 3)]

expected_r = [np.array((1.41774469)).reshape(1, 1, 1),
              np.array((2.44948974)).reshape(1, 1, 1)]

expected_theta = [np.array((0.78788574)).reshape(1, 1, 1),
                  np.array((1.15026199)).reshape(1, 1, 1)]

expected_phi = [np.array((1.47112767)).reshape(1, 1, 1),
                np.array((1.10714872)).reshape(1, 1, 1)]

expected_hash = ['109467', '220712']


@pytest.mark.parametrize('vector', [v for v in vectors])
def test_vector(vector):
    pc = ProbedCoordinates(vector)
    assert np.allclose(pc.vector, vector)
    pc = ProbedCoordinates
    pc.vector = vector
    assert np.allclose(pc.vector, vector)


@pytest.mark.parametrize('vector, exp_r, exp_theta, exp_phi', [t for t in zip(vectors,
                                                                              expected_r,
                                                                              expected_theta,
                                                                              expected_phi)])
def test_to_spherical(vector, exp_r, exp_theta, exp_phi):
    pc = ProbedCoordinates(vector)
    r, theta, phi = pc.to_spherical
    assert np.allclose(r, exp_r)
    assert np.allclose(theta, exp_theta)
    assert np.allclose(phi, exp_phi)


@pytest.mark.parametrize('vector, expected_hash', [t for t in zip(vectors, expected_hash)])
def test_hash(vector, expected_hash):
    pc = ProbedCoordinates(vector)
    assert expected_hash == str(hash(pc))[:6]
    # Should be rounded off by hash calculation
    pc.vector = vector + abs(vector) * 1e-8
    assert expected_hash == str(hash(pc))[:6]
    # Strict equality should fail
    assert np.all(vector != pc.vector)
    # Approximate equality should hold
    assert np.allclose(vector, pc.vector)
    pc.vector = vector + vector.min() * 1e-4
    assert str(hash(pc))[:6] != expected_hash



import pytest  # noqa:F401
import numpy as np

from mumott.core import cuda_utils

dtype_list = [np.int32, np.float32, np.float64, np.int64]
shape_list = [(3, 3, 3) for _ in dtype_list]


@pytest.mark.parametrize('shape, dtype', [t for t in zip(shape_list, dtype_list)])
def test_cuda_calloc(shape, dtype):
    a = cuda_utils.cuda_calloc(shape, dtype)
    assert np.allclose(0, a.copy_to_host())
    assert np.allclose(shape, a.shape)
    if dtype == np.float64:
        assert a.dtype == np.float32
    else:
        assert a.dtype == dtype



import pytest
import logging

import numpy as np
from mumott import DataContainer

from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjectorNumba
from mumott.methods.functionals import SIGTT

LOGGER = logging.getLogger(__name__)

new_pc = np.array([[[0.93896659,  0.04472849, -0.34108812],
                   [0.9655966,   0.04599704, -0.25594428],
                   [0.98454023,  0.04689943, -0.16876307],
                   [0.99564669,  0.0474285, -0.08023846],
                   [0.99882756,  0.04758002,  0.00892486],
                   [0.99405752,  0.0473528,   0.09801714],
                   [0.98137455,  0.04674863,  0.18632918],
                   [0.9608796,   0.04577234,  0.273158],
                   [0.93273582,  0.04443168,  0.35781241],
                   [0.89716724,  0.04273734,  0.43961855],
                   [0.85445699,  0.0407028,   0.51792522]],
                  [[0.7996345,   0.03809129,  0.59927767],
                   [0.73106553,  0.03482494,  0.68141794],
                   [0.654119,    0.03115953,  0.75574957],
                   [0.56967665,  0.02713704,  0.82142078],
                   [0.47870616,  0.02280358,  0.87767899],
                   [0.38224998,  0.01820881,  0.92387953],
                   [0.28141345,  0.01340537,  0.95949297],
                   [0.17735209,  0.00844832,  0.9841112],
                   [0.07125838,  0.00339446,  0.99745211],
                   [-0.03565191, -0.00169831,  0.99936283],
                   [-0.14215364, -0.00677161,  0.98982144]],
                  [[-0.23837871, -0.01135538,  0.97110589],
                   [-0.32388788, -0.01542868,  0.94596966],
                   [-0.40681882, -0.01937917,  0.91330329],
                   [-0.48651138, -0.02317539,  0.8733668],
                   [-0.5623312, -0.02678714,  0.82647811],
                   [-0.63367473, -0.03018565,  0.77301045],
                   [-0.69997406, -0.03334387,  0.71338945],
                   [-0.76070142, -0.03623667,  0.64808969],
                   [-0.81537342, -0.03884102,  0.57763099],
                   [-0.86355485, -0.04113619,  0.5025742],
                   [-0.90486218, -0.0431039,   0.42351681]]])


@pytest.mark.parametrize('vec', [(np.array((0.2, 0.3, 0.7)))])
def test_probed_coordinates(vec):
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    old_pc = bs.probed_coordinates.vector
    meth = SIGTT(dc, bs, pr)  # noqa
    dc.geometry.basis_vector_p = vec
    print(bs.probed_coordinates.vector)
    assert not np.allclose(old_pc, bs.probed_coordinates.vector)
    assert np.allclose(new_pc, bs.probed_coordinates.vector)
    assert np.allclose(bs.probed_coordinates.vector, meth.probed_coordinates)


def test_is_dirty():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    old_hash = hash(meth)
    assert not meth.is_dirty
    dc.geometry.p_direction_0 = np.array((0, 0, 1))
    assert old_hash != hash(meth)
    old_hash = hash(meth)
    assert meth.is_dirty
    meth._update()
    assert old_hash != hash(meth)
    assert not meth.is_dirty


new_coeffs = [np.ones((4, 4, 4, 1)),
              np.ones((4, 4, 2, 1))]


@pytest.mark.parametrize('new_coeff, valid, myhash', [(new_coeffs[0], True, '184781'),
                                                      (new_coeffs[1], False, None)])
def test_coeff(new_coeff, valid, myhash):
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    if valid:
        old_hash = hash(meth)
        meth.coefficients = new_coeff
        assert old_hash != hash(meth)
        assert myhash == str(hash(meth))[:6]
        assert np.allclose(meth.coefficients, new_coeff)
    else:
        with pytest.raises(ValueError):
            meth.coefficients = new_coeff


def test_get_residuals():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    old_hash = hash(meth)
    out = meth.get_residuals(True)
    residuals, gradient = (out['residuals'], out['gradient'])
    assert old_hash == hash(meth)
    val = np.array([-12.88, -14.4, -21.55])
    assert np.allclose(residuals[0, 0, 2], val)
    grad_slice = np.array([[-47.51],
                          [-47.77],
                          [-48.83],
                          [-50.04]])
    assert np.allclose(meth.get_residuals()['residuals'][0, 0, 2], val)
    assert np.allclose(grad_slice, gradient[0, 1])


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    s = str(meth)
    assert '225ef2' in s


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    h = meth._repr_html_()
    assert '225ef2' in h


def test_coeff_update(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    dc.geometry.volume_shape = np.array((5, 4, 4))
    meth._update()
    assert 'truncated' in caplog.text
    assert np.allclose(meth.coefficients.shape[:-1], np.array((5, 4, 4)))
    caplog.clear()
    bs.ell_max = 2
    meth._update()
    assert 'truncated' in caplog.text
    assert 'Size of basis set' in caplog.text
    caplog.clear()
    assert np.allclose(meth.coefficients.shape, np.array((5, 4, 4, 6)))
    bs.ell_max = 0
    meth._update()
    assert 'truncated' in caplog.text
    assert np.allclose(meth.coefficients.shape, np.array((5, 4, 4, 1)))



from typing import Dict

import numpy as np
from numpy.typing import NDArray

from ..regularizers.base_regularizer import Regularizer


class Laplacian(Regularizer):

    """Regularizes using the Laplacian of the coefficients. Suitable for orthonormal representations,
    e.g., spherical harmonics.
    """

    def __init__(self):
        super().__init__()

    def get_regularization_norm(self,
                                coefficients: NDArray[float],
                                get_gradient: bool = False) -> Dict:
        """Retrieves the regularization norm and possibly the gradient based on the provided coefficients.
        The norm is the 2-norm of the discrete nearest-neighbour Laplacian.

        This is in effect a smoothing kernel that enforces continuity between the tensors of
        neighbouring voxels. The calculation is most suitable for orthonormal representations.
        If the representation is in spherical harmonics, the norm corresponds to maximimzing
        the covariance between neighbours.

        Parameters
        ----------
        coefficients
            An ``np.ndarray`` of values, with shape ``(X, Y, Z, W)``, where
            the last channel contains, e.g., tensor components.
        get_gradient
            If ``True``, returns a ``'gradient'`` of the same shape as :attr:`coefficients`.
            Otherwise, the entry ``'gradient'`` will be ``None``. Defaults to ``False``.

        Returns
        -------
            A dictionary with two entries, ``regularization_norm`` and ``gradient``.
        """
        r = 6 * coefficients
        slices_r = [np.s_[1:, :, :], np.s_[:-1, :, :],
                    np.s_[:, 1:, :], np.s_[:, :-1, :],
                    np.s_[:, :, 1:], np.s_[:, :, :-1]]
        slices_coeffs = [np.s_[:-1, :, :], np.s_[1:, :, :],
                         np.s_[:, :-1, :], np.s_[:, 1:, :],
                         np.s_[:, :, :-1], np.s_[:, :, 1:]]

        for s, v in zip(slices_r, slices_coeffs):
            r[s] -= coefficients[v]

        r *= 0.5

        result = dict(regularization_norm=None, gradient=None)

        if get_gradient:
            result['gradient'] = r

        result['regularization_norm'] = np.dot(r.ravel(), r.ravel())

        return result

    @property
    def _function_as_str(self) -> str:
        return 'R(x) = 0.5 * lambda * (2 * x[i] - x[i + 1] - x[i - 1]) ** 2'

    @property
    def _function_as_tex(self) -> str:
        return r'$R(\vec{x}) = \lambda \frac{\Vert \nabla^2 \vec{x} \Vert^2}{2}$'



# -*- coding: utf-8 -*-


from .john_transform import john_transform, john_transform_adjoint
from .john_transform_cuda import john_transform_cuda, john_transform_adjoint_cuda
from .cuda_utils import cuda_calloc
from .spherical_harmonic_mapper import SphericalHarmonicMapper


__all__ = [
    'john_transform_adjoint',
    'john_transform_adjoint_cuda',
    'john_transform',
    'john_transform_cuda',
    'cuda_calloc',
    'SphericalHarmonicMapper'
]



from dataclasses import dataclass
from typing import Tuple
import numpy as np

from numpy.typing import NDArray

from .hashing import list_to_hash


@dataclass
class ProbedCoordinates:
    """ A small container class for probed coordinates for the
    :class:` BasisSet <mumott.methods.BasisSet>` being used
    together with the :class:`Method <Method>`.

    Parameters
    ----------
    vector : NDArray[float]
        The coordinates on the sphere probed at each detector segment by the
        experimental method. Should be structured ``(N, M, I, 3)`` where ``N``
        is the number of projections, ``M`` is the number of detector segments,
        ``I`` is the number of points on the detector to be integrated over,
        and the last index gives the ``(x, y, z)`` components of the coordinates.
        ``I`` can be as small as ``1``, but the array still needs to be structured
        in this way explicitly.
        By default, the value will be ``np.array([1., 0., 0.]).reshape(1, 1, 1, 3)``.

    """
    vector: NDArray[float] = np.array([1., 0., 0.]).reshape(1, 1, 1, 3)

    def __hash__(self) -> int:
        return int(list_to_hash([self.vector]), 16)

    @property
    def to_spherical(self) -> Tuple:
        """ Returns spherical coordinates of :attr:`vector`, in the order
        ``(radius, polar angle, azimuthal angle)``. """
        r = np.linalg.norm(self.vector, axis=-1)
        theta = np.arccos(self.vector[..., 2] / r)
        phi = np.arctan2(self.vector[..., 1], self.vector[..., 0])
        return (r, theta, phi)



import logging
from typing import Dict

import numpy as np
from numpy.typing import NDArray

from .base_loss_function import LossFunction
from ...methods.functionals.base_functional import Functional

logger = logging.getLogger(__name__)


class SquaredLoss(LossFunction):

    r"""Class object for obtaining the squared loss function and
    gradient from a given :attr:`functional`.

    This loss function can be written as :math:`L(r(x, d)) = 0.5 r(x, d)^2`, where :math:`r` is the
    residual, a function of :math:`x`, the optimization coefficients, and :math:`d`, the data.
    The gradient with respect to :math:`x` is then :math:`\frac{\partial r}{\partial x}`.
    The partial derivative of :math:`r` with respect to :math:`x` is the responsibility
    of the :attr:`functional` to compute.

    Generally speaking, the squared loss function is easy to compute and has a well-behaved
    gradient, but it is not robust against outliers in the data. Using weights to normalize
    residuals by the variance can mitigate this somewhat.

    Parameters
    ----------
    functional : :class:`Functional`
        The functional instance from which the residuals, weights and gradient terms
        are obtained.
    use_weights : bool
        Whether to use weighting in the computation of the residual norm and gradient.
        Default is ``False``.
    preconditioner : np.ndarray
        A preconditioner to be applied to the gradient. Must have the same shape as
        :attr:`functional.coefficients <mumott.methods.functionals.base_functional.Functional.coefficients>`
        or it must be possible to broadcast by multiplication.
    residual_norm_multiplier : float
        A multiplier that is applied to the residual norm and gradient. Useful in cases where
        a very small or large loss function value changes the optimizer behaviour.
    """

    def __init__(self,
                 functional: Functional,
                 use_weights: bool = False,
                 preconditioner: NDArray[float] = None,
                 residual_norm_multiplier: float = 1):
        super().__init__(functional, use_weights, preconditioner, residual_norm_multiplier)

    def _get_residual_norm_internal(self, get_gradient: bool = False) -> Dict:
        """ Gets the residual norm, and if needed,
        the gradient, using the attached :attr:`functional`.

        Parameters
        ----------
        get_gradient
            Whether to return the gradient. Default is ``False``.

        Returns
        -------
             A ``dict`` with two entries, ``residual_norm`` and ``gradient``.
             If ``get_gradient`` is false, its value will be ``None``.
        """
        functional_output = self._functional.get_residuals(
            get_gradient=get_gradient, get_weights=self._use_weights)
        residuals = functional_output['residuals']
        if self.use_weights:
            # weights (1/variance) need to be applied since they depend on the loss function
            residual_norm = 0.5 * np.einsum(
                'ijkh, ijkh, ijkh -> ...', residuals, residuals, functional_output['weights'])
        else:
            residual_norm = 0.5 * np.einsum(
                'ijkh, ijkh -> ...', residuals, residuals)

        if residual_norm < 1:
            logger.warning(f'The residual norm value ({residual_norm}) is < 1.'
                           ' Note that some optimizers change their convergence criteria for'
                           ' loss functions < 1!')

        return dict(residual_norm=residual_norm, gradient=functional_output['gradient'])

    @property
    def _function_as_str(self) -> str:
        """ Should return a string representation of the associated loss function. """
        return 'L(r) = r ** 2'

    @property
    def _function_as_tex(self) -> str:
        """ Should return a string representation of the associated loss function
        in MathJax-renderable TeX."""
        return r'$L(r) = r^2$'



""" Container for class OrientationImageMapper. """
from matplotlib.colors import Colormap
from matplotlib.cm import get_cmap
import colorcet # noqa
from typing import Union
import numpy as np


class OrientationImageMapper():

    def __init__(self,
                 colormap: Union[str, Colormap] = 'cet_cyclic_isoluminant'):
        """
        Helper class for generating a color wheel to be used in plots.

        Parameters
        ----------
        colormap : Union[str, Colormap]
            The colormap to be used for creating the colorwheel. Should be cyclic.
            Default is ``cet_cyclic_isoluminant``.
        """
        self._colormap = get_cmap(colormap)
        r = np.sqrt(np.linspace(1.0, 0.0, 360))
        th = np.linspace(0, 2 * np.pi, 360)
        th, r = np.meshgrid(th, r)
        th_flip = np.arctan2(np.cos(th), np.sin(th))
        thx = (th[:-1, :-1] - (th[:-1, :-1] > np.pi).astype(float) * np.pi) / np.pi
        thx = self._colormap(thx)
        self._wheel_properties = [th_flip, r, thx]

    @property
    def colormap(self) -> Colormap:
        """ Returns the colormap used by the wheel. """
        return self._colormap

    @property
    def wheel_properties(self) -> list:
        """ Returns wheel properties as a list. The entry at
        ``wheel_properties[0]`` contains the angles. ``wheel_properties[1]`` contains the radius.
        ``wheel_properties[2]`` contains the RGBA values mapped to the angles. """
        return self._wheel_properties



import logging

from typing import Tuple

import numpy as np
from numpy.typing import NDArray
from numba.cuda import is_cuda_array

from ...core.cuda_utils import cuda_calloc
from ...core.john_transform_cuda import john_transform_cuda, john_transform_adjoint_cuda
from ...core.hashing import list_to_hash
from ... import Geometry
from .base_projector import Projector

logger = logging.getLogger(__name__)


class SAXSProjectorCUDA(Projector):
    """
    Projector for transforms of tensor fields from three-dimensional space
    to projection space. Uses a projection algorithm implemented in
    ``numba.cuda``.

    Parameters
    ----------
    geometry : Geometry
        An instance of :class:`Geometry <mumott.Geometry>` containing the
        necessary vectors to compute forwared and adjoint projections.
    """
    def __init__(self,
                 geometry: Geometry):

        super().__init__(geometry)

        self._basis_vector_projection = np.zeros((len(self._geometry), 3), dtype=np.float64)
        self._basis_vector_j = np.zeros((len(self._geometry), 3), dtype=np.float64)
        self._basis_vector_k = np.zeros((len(self._geometry), 3), dtype=np.float64)
        self._update(force_update=True)

    def _get_john_transform_parameters(self,
                                       indices: NDArray[int] = None) -> Tuple:
        if indices is None:
            indices = np.arange(len(self._geometry))
        vector_p = self._basis_vector_projection[indices]
        vector_j = self._basis_vector_j[indices]
        vector_k = self._basis_vector_k[indices]
        j_offsets = self._geometry.j_offsets_as_array[indices]
        k_offsets = self._geometry.k_offsets_as_array[indices]
        return (vector_p, vector_j, vector_k, j_offsets, k_offsets)

    @staticmethod
    def _get_zeros_method(array: NDArray):
        """ Internal method for returning `cupy` or `numpy` based on the type of array given. """
        if is_cuda_array(array):
            return cuda_calloc
        else:
            return np.zeros

    def forward(self,
                field: NDArray[np.float32],
                indices: NDArray[int] = None) -> NDArray:
        """ Compute the forward projection of a tensor field.

        Parameters
        ----------
        field
            An array containing coefficients in its fourth dimension,
            which are to be projected into two dimensions. The first three
            dimensions should match the :attr:`volume_shape` of the sample.
            Can be either a ``numpy.ndarray`` or any array that implements
            the CUDA array interface. Must be single-precision.
        indices
            A one-dimensional array containing one or more indices
            indicating which projections are to be computed. If ``None``,
            all projections will be computed.

        Returns
        -------
            An array with four dimensions ``(I, J, K, L)``, where
            the first dimension matches :attr:`indices`, such that
            ``projection[i]`` corresponds to the geometry of projection
            ``indices[i]``. The second and third dimension contain
            the pixels in the ``J`` and ``K`` dimension respectively, whereas
            the last dimension is the coefficient dimension, matching ``field[-1]``.
        """
        assert np.allclose(field.shape[:-1], self._geometry.volume_shape)
        self._update()
        if indices is None:
            return self._forward_stack(field)
        return self._forward_subset(field, indices)

    def _forward_subset(self,
                        field: NDArray[np.float32],
                        indices: NDArray[int]) -> NDArray:
        """ Internal method for computing a subset of projections.

        Parameters
        ----------
        field
            The field to be projected. Must be single-precision.
        indices
            The indices indicating the subset of all projections in the
            system geometry to be computed.

        Returns
        -------
            The resulting projections.
        """
        indices = np.array(indices).ravel()
        self._check_indices_kind_is_integer(indices)
        init_method = self._get_zeros_method(field)
        projections = init_method((indices.size,) +
                                  tuple(self._geometry.projection_shape) +
                                  (field.shape[-1],), dtype=np.float32)
        john_transform_cuda(field, projections, *self._get_john_transform_parameters(indices))
        return projections

    def _forward_stack(self,
                       field: NDArray[np.float32]) -> NDArray:
        """Internal method for forward projecting an entire stack.

        Parameters
        ----------
        field
            The field to be projected. Must be single-precision.

        Returns
        -------
            The resulting projections.
        """
        init_method = self._get_zeros_method(field)
        projections = init_method((len(self._geometry),) +
                                  tuple(self._geometry.projection_shape) +
                                  (field.shape[-1],), dtype=np.float32)
        john_transform_cuda(field, projections, *self._get_john_transform_parameters())
        return projections

    def adjoint(self,
                projections: NDArray[np.float32],
                indices: NDArray[int] = None) -> NDArray:
        """ Compute the adjoint of a set of projections according to the system geometry.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from, e.g., the residual of measured data and forward projections.
            The first dimension should match :attr:`indices` in size, and the
            second and third dimensions should match the system projection geometry.
            Can be either a ``numpy.ndarray`` or any array that implements
            the CUDA array interface. Must be single-precision.
        indices
            A one-dimensional array containing one or more indices
            indicating from which projections the adjoint is to be computed.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial and the last dimension runs over
            coefficients.
        """
        assert np.allclose(projections.shape[-3:-1], self._geometry.projection_shape)
        self._update()
        if indices is None:
            return self._adjoint_stack(projections)
        return self._adjoint_subset(projections, indices)

    def _adjoint_subset(self,
                        projections: NDArray[np.float32],
                        indices: NDArray[int]) -> NDArray:
        """ Internal method for computing the adjoint of only a subset of projections.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from, e.g., the residual of measured data and forward projections.
            The first dimension should match :attr:`indices` in size, and the
            second and third dimensions should match the system projection geometry.
            Must be single-precision.
        indices
            A one-dimensional array containing one or more indices
            indicating from which projections the adjoint is to be computed.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial and the last dimension runs over
            coefficients. """
        if projections.ndim == 3:
            assert indices.size == 1
            projections = projections[np.newaxis, ...]
        else:
            assert indices.size == projections.shape[0]
        indices = np.array(indices).ravel()
        self._check_indices_kind_is_integer(indices)
        init_method = self._get_zeros_method(projections)
        field = init_method(tuple(self._geometry.volume_shape) +
                            (projections.shape[-1],), dtype=np.float32)
        # Note that we assume projections match ``indices`` in their layout.
        john_transform_adjoint_cuda(field,
                                    projections[indices],
                                    *self._get_john_transform_parameters(indices))
        return field

    def _adjoint_stack(self,
                       projections: NDArray[np.float32]) -> NDArray:
        """ Internal method for computing the adjoint of a whole stack of projections.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from e.g. the residual of measured data and forward projections.
            The first dimension should run over all the projection directions
            in the system geometry. Must be single-precision.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial, and the last dimension runs over
            coefficients. """
        assert projections.shape[0] == len(self._geometry)
        init_method = self._get_zeros_method(projections)
        field = init_method(tuple(self._geometry.volume_shape) +
                            (projections.shape[-1],), dtype=np.float32)
        john_transform_adjoint_cuda(field, projections, *self._get_john_transform_parameters())
        return field

    def _update(self,
                force_update: bool = False) -> None:
        super()._update(force_update)
        self._calculate_basis_vectors()
        self._geometry_hash = hash(self._geometry)

    def _calculate_basis_vectors(self) -> None:
        """ Calculates the basis vectors for the John transform, one projection vector
        and two coordinate vectors. """
        self._basis_vector_projection = np.einsum(
            'kij,i->kj', self._geometry.rotations_as_array, self._geometry.p_direction_0)
        self._basis_vector_j = np.einsum(
            'kij,i->kj', self._geometry.rotations_as_array, self._geometry.j_direction_0)
        self._basis_vector_k = np.einsum(
            'kij,i->kj', self._geometry.rotations_as_array, self._geometry.k_direction_0)

    def __hash__(self) -> int:
        to_hash = [self._basis_vector_projection,
                   self._basis_vector_j,
                   self._basis_vector_k,
                   self._geometry_hash,
                   hash(self._geometry)]
        return int(list_to_hash(to_hash), 16)

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += [self.__class__.__name__.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, edgeitems=2, precision=5, linewidth=60):
            s += ['{:18} : {}'.format('is_dirty', self.is_dirty)]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += [f'<h3>{self.__class__.__name__}</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">is_dirty</td>']
            s += [f'<td>1</td><td>{self.is_dirty}</td></tr>']
            s += ['<tr><td style="text-align: left;">hash</td>']
            s += [f'<td>{len(hex(hash(self)))}</td><td>{hex(hash(self))[2:8]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)



""" Container for class SphericalHarmonicMapper. """
import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation as rot
from scipy.special import sph_harm, factorial2
from typing import Tuple


class SphericalHarmonicMapper:
    """
    Helper class for visualizing and analyzing spherical harmonics.
    Using this class, one can obtain the amplitudes for a given set
    of even-ordered harmonics and apply the Funk-Radon transform.
    These can then be plotted, analyzed, and so forth. In addition, the class
    allows one to represent functions in terms of spherical harmonics,
    if they are given as functions of the azimuthal and polar angles
    of the class instance.

    Parameters
    ----------
    max_order : int, optional
        Maximum order of the spherical harmonics. Default is ``2``.
    polar_resolution : int, optional
        Number of samples in the polar direction. Default is ``16``.
    azimuthal_resolution : int, optional
        Number of samples in the azimuthal direction. Default is ``32``.
    polar_zero : float, optional
        The polar angle of the spherical harmonic coordinate system's
        pole, relative to the reference coordinate system. Default is ``0``.
    azimuthal_zero : float, optional
        The azimuthal angle of the spherical harmonic coordinate system's
        pole, relative to a reference coordinate system. Default is ``0``.

    Example
    -------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> S = SphericalHarmonicMapper(max_order=4, polar_resolution=16, azimuthal_resolution=8)
        >>> S.orders
        array([0, 2, 2, 2, 2, 2, 4, 4, ...])
        >>> S.degrees
        array([ 0, -2, -1,  0,  1,  2, -4, -3, ...])
        >>> a = S.get_amplitudes(np.random.rand(S.orders.size) - 0.5)
        >>> plt.pcolormesh(S.phi * np.sqrt(1 / 2), # basic cylindrical equal-area projection
                           np.cos(S.theta) / np.sqrt(1 / 2),
                           a[:-1, :-1])
        ...
    """

    def __init__(self,
                 max_order: int = 2,
                 polar_resolution: int = 16,
                 azimuthal_resolution: int = 32,
                 polar_zero: float = 0,
                 azimuthal_zero: float = 0):
        self._polar_zero = polar_zero
        self._azimuthal_zero = azimuthal_zero
        self._max_order = max_order
        polar_coordinates = np.linspace(0, np.pi, polar_resolution)
        azimuthal_coordinates = np.linspace(-np.pi, np.pi, azimuthal_resolution + 1)[:-1]
        self._theta, self._phi = np.meshgrid(
            polar_coordinates,
            azimuthal_coordinates,
            indexing='ij')
        self._update_l_and_m()

        self._map = np.zeros(self._theta.shape + self._l.shape)
        self._complex_map = np.zeros(self._theta.shape + self._l.shape).astype(complex)
        self._update_map()

        self._rotated_system = True
        self._get_coordinates()
        self._update_funk_coefficients()

        self._xyz = np.concatenate((self._X[..., None], self._Y[..., None], self._Z[..., None]), axis=2)
        self.update_zeros(self._polar_zero, self._azimuthal_zero)

    def get_amplitudes(self,
                       coefficients: NDArray[float],
                       apply_funk_transform: bool = False) -> NDArray[float]:
        """
        Returns the amplitudes of a set of spherical harmonics. For sorting
        of the coefficients, see the :attr:`orders` and :attr:`degrees` attributes.

        Parameters
        ----------
        coefficients
            The coefficients for which the amplitudes are to be calculated.
            Size must be equal to ``(max_order + 1) * (max_order / 2 + 1)``.
        apply_funk_fransform
            Whether to apply the Funk-Radon transform to the coefficients.
            This is useful for orientation analysis in some cases.
            Default is ``False``.
        """
        if apply_funk_transform:
            coefficients = (self._funk_coefficients * coefficients.ravel()).reshape(1, 1, -1)
        else:
            coefficients = coefficients.reshape(1, 1, -1)
        return np.einsum('...i, ...i', coefficients, self._map)

    def get_harmonic_coefficients(self, amplitude: NDArray[float]) -> NDArray[float]:
        """ Returns the spherical harmonic coefficients for the given amplitudes.
        Can be used with amplitudes from another instance of
        :class:`SphericalHarmonicParameters` with a different orientation
        in order to solve for the rotated coefficients. The accuracy of the
        representation depends on the maximum order and the polar and azimuthal
        resolution.

        Parameters
        ----------
        amplitude
            The amplitude of the spherical function to be
            represented, as a function of ``theta_interpolator``
            and ``phi_interpolator``.
        """
        area_normer = np.sin(self.theta_interpolator)
        area_normer /= np.sum(area_normer)
        scaled_amp = np.einsum('..., ...',
                               amplitude,
                               area_normer).reshape(self.theta_interpolator.shape + (1,))
        coeffs = np.einsum('ij..., ij...', self.map_interpolator, scaled_amp)
        return coeffs

    def _get_coordinates(self) -> Tuple[NDArray[float], NDArray[float], NDArray[float]]:
        """ Gets the X, Y, and Z-coordinates. Updates them only if the system has
        been rotated since the last call. """
        if self._rotated_system:
            self._X, self._Y, self._Z = \
                (np.multiply(0.5, np.einsum('..., ...', np.sin(self._theta), np.cos(self._phi))),
                 np.multiply(0.5, np.einsum('..., ...', np.sin(self._theta), np.sin(self._phi))),
                 np.multiply(0.5, np.cos(self._theta)))
            self._rotated_system = False
        return self._X, self._Y, self._Z

    def _update_funk_coefficients(self) -> None:
        """ Updates the Funk coefficients used for the Funk transform. """
        funk_coefficients = []
        for i in range(self._max_order+1):
            if i % 2 == 0:
                funk_coefficients.append(((-1) ** (i // 2)) * factorial2(i - 1) / factorial2(i))
        funk_coefficients = np.array(funk_coefficients)
        self._funk_coefficients = funk_coefficients[self._l // 2]

    def _update_l_and_m(self) -> None:
        """ Updates the order and degree vectors of the system. """
        self._l = []
        self._m = []
        for ll in range(0, self._max_order+1, 2):
            for mm in range(-ll, ll+1):
                self._l.append(ll)
                self._m.append(mm)
        self._l = np.array(self._l)
        self._m = np.array(self._m)

    def _update_map(self) -> None:
        """ Updates the map of the system. """
        self._update_complex_map()
        self._update_real_map()

    def _update_complex_map(self) -> None:
        """ Retrieves the complex map used for determining
        the real map. """
        self._complex_map[...] = sph_harm(
            abs(self._m.reshape(1, 1, -1)),
            self._l.reshape(1, 1, -1),
            self._phi[:, :, None],
            self._theta[:, :, None])

    def _update_real_map(self) -> None:
        """ Calculates the real spherical harmonic map based on
        the complex map. """
        self._map[:, :, self._m == 0] = \
            np.sqrt(4 * np.pi) * self._complex_map[:, :, self._m == 0].real
        self._map[:, :, self._m > 0] = \
            ((-1.) ** (self._m[self._m > 0])) * np.sqrt(2) * \
            np.sqrt(4 * np.pi) * self._complex_map[:, :, self._m > 0].real
        self._map[:, :, self._m < 0] = \
            ((-1.) ** (self._m[self._m < 0])) * np.sqrt(2) * \
            np.sqrt(4 * np.pi) * self._complex_map[:, :, self._m < 0].imag

    def update_zeros(self, polar_zero: float, azimuthal_zero: float) -> None:
        """Changes the orientation of the coordinate system.

        Parameters
        ----------
        polar_zero
            The new polar angle at which the pole should be,
            relative to a fixed reference system.
        azimuthal_zero
            The new azimuthal angle at which the pole should be,
            relative to a fixed reference system.
        """
        if polar_zero == 0 and azimuthal_zero == 0:
            self._theta = np.arccos(self._xyz[..., 2] / np.sqrt((self._xyz ** 2).sum(-1)))
            self._phi = np.arctan2(self._xyz[..., 1], self._xyz[..., 0])
            return
        xyz = np.copy(self._xyz)
        xyz = xyz.reshape(-1, 3)
        rotvec = rot.from_euler('Z', (-azimuthal_zero))
        rotvec = rot.from_euler('Y', (-polar_zero)) * rotvec
        xyz = rotvec.apply(xyz).reshape(self._theta.shape + (3,))
        self._theta = np.arccos(xyz[..., 2] / np.sqrt((xyz ** 2).sum(-1)))
        self._phi = np.arctan2(xyz[..., 1], xyz[..., 0])
        self._polar_zero = polar_zero
        self._azimuthal_zero = azimuthal_zero
        self._rotated_system = True
        self._update_map()

    @property
    def orders(self) -> NDArray[int]:
        """ The orders of the harmonics calculated
        by the class instance. """
        return self._l

    @property
    def degrees(self) -> NDArray[int]:
        """ The degrees of the harmonics calculated
        by the class instance. """
        return self._m

    @property
    def polar_zero(self) -> NDArray[float]:
        """ The polar angle of the spherical harmonic pole,
        relative to a fixed reference system. """
        return self._polar_zero

    @property
    def azimuthal_zero(self) -> NDArray[float]:
        """ The azimuthal angle of the spherical harmonic pole,
        relative to a fixed reference system. """
        return self._azimuthal_zero

    @property
    def theta(self) -> NDArray[float]:
        """ The polar angle to which the amplitude is mapped. """
        return self._theta

    @property
    def phi(self) -> NDArray[float]:
        """ The azimuthal angle to which the amplitude is mapped. """
        return self._phi

    @property
    def phi_interpolator(self) -> NDArray[float]:
        """ The azimuthal angle with an end-point removed,
        to be used for solving for spherical harmonics. """
        return self._phi[:, :-1]

    @property
    def theta_interpolator(self) -> NDArray[float]:
        """ The polar angle with an end-point removed,
        to be used for solving for spherical harmonics. """
        return self._theta[:, :-1]

    @property
    def max_order(self) -> int:
        """ Maximum order of the spherical harmonics. """
        return self._max_order

    @property
    def map(self) -> NDArray[float]:
        """ Map between amplitude and harmonics. """
        return self._map

    @property
    def map_interpolator(self) -> NDArray[float]:
        """ Maximum order of the spherical harmonics. """
        return self._map[:, :-1]

    @property
    def coordinates(self) -> Tuple[NDArray[float], NDArray[float], NDArray[float]]:
        """ The X, Y, Z coordinates that the amplitudes
        are mapped to. """
        return self._get_coordinates()

    @max_order.setter
    def max_order(self, new_max_order: int):
        self._max_order = new_max_order
        self._update_l_and_m()
        self._map = np.zeros(self._theta.shape + self._l.shape)
        self._complex_map = np.zeros(self._theta.shape + self._l.shape).astype(complex)
        self._update_map()
        self._update_funk_coefficients()



import pytest # noqa
import numpy as np
from mumott.data_handling import DataContainer


def test_load():
    dc = DataContainer('tests/test_half_circle.h5')
    assert np.allclose(dc.data.shape, (1, 4, 4, 3))
    assert np.allclose(dc.weights.shape, (1, 4, 4, 3))
    old_gm = dc.geometry
    old_frame = dc.stack[0]
    dc = DataContainer('tests/test_half_circle.h5', skip_data=True)
    assert np.allclose(dc.diode.shape, (1, 4, 4))
    assert np.allclose(dc.weights.shape, (1, 4, 4))
    assert np.allclose(dc.geometry[0].rotation, old_gm[0].rotation)
    assert np.allclose(dc.geometry[0].j_offset, old_gm[0].j_offset)
    assert np.allclose(dc.geometry[0].k_offset, old_gm[0].k_offset)
    assert hash(dc.geometry) != hash(old_gm)
    old_gm.projection_shape = dc.geometry.projection_shape
    old_gm.detector_angles = dc.geometry.detector_angles
    assert hash(dc.geometry) == hash(old_gm)
    assert np.all(dc.stack[0].data == np.array([]).reshape(0, 0))
    assert np.allclose(dc.stack[0].diode, old_frame.diode)
    assert np.allclose(dc.stack[0].weights, old_frame.weights[..., 0])


new_diode = np.arange(16).reshape(1, 4, 4)
new_diode_list = [new_diode[0].copy()]


def test_set_diode():
    dc = DataContainer('tests/test_half_circle.h5')
    dc.stack.diode = new_diode
    assert np.allclose(dc.diode, new_diode)
    assert np.allclose(dc.stack[0].diode, new_diode[0])
    dc.stack.diode = new_diode_list
    assert np.allclose(dc.diode, new_diode_list)
    assert np.allclose(dc.stack[0].diode, new_diode_list[0])


def test_set_weights():
    dc = DataContainer('tests/test_half_circle.h5', skip_data=True)
    dc.stack.weights = new_diode
    assert np.allclose(dc.weights, new_diode)
    assert np.allclose(dc.stack[0].weights, new_diode[0])
    dc.stack.weights = new_diode_list
    assert np.allclose(dc.weights, new_diode_list)
    assert np.allclose(dc.stack[0].weights, new_diode_list[0])


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    s = str(dc)
    assert s.count('\n') <= 50


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    s = dc._repr_html_()
    assert s.count('</tr>') <= 25


def test_correct_for_transmission():
    dc = DataContainer('tests/test_half_circle.h5')
    dc.stack.diode = dc.diode * 0.5
    data = dc.stack.data
    dc.correct_for_transmission()
    assert np.allclose(data * 2, dc.stack.data)
    assert dc._correct_for_transmission_called


def test_get_str():
    dc = DataContainer('tests/test_half_circle.h5')
    s = dc._get_str_representation(5)
    assert s.count('\n') <= 6


def test_get_html():
    dc = DataContainer('tests/test_half_circle.h5')
    s = dc._get_html_representation(5)
    assert s.count('</tr>') <= 6



import logging

from typing import Tuple

from numba import cuda
from numpy.typing import DTypeLike, NDArray

from numpy import float32, float64

logger = logging.getLogger(__name__)


def cuda_calloc(shape: Tuple, dtype: DTypeLike = float32) -> NDArray:
    """ Function for creating a zero-initialized
    device array using ``numba.cuda``, as it provides no
    native method for doing so.

    Paramters
    ---------
    shape
        A ``tuple`` with the shape of the desired
        device array.
    dtype
        A ``numba`` or ``numpy`` dtype. ``float64``
        is turned into ``float32``.
    """
    if dtype == float64:
        dtype = float32
    array = cuda.device_array(shape, dtype=dtype)

    size = array.size

    @cuda.jit
    def _calloc(array: NDArray):
        """ Initializes the given array. Depends on
        `size` being predefined.

        Parameters
        ----------
        array
            A device array which is initialized with zeros.
        """
        x = cuda.grid(1)
        if x < size:
            array[x] = 0

    # Use 512 threads per block, and blocks per grid based on array size.
    tpb = 512
    bpg = (size + tpb - 1) // tpb
    _calloc[bpg, tpb](array.ravel())
    return array
